from typing import Optional, Union

from fractions import Fraction

from musurgia.musurgia_exceptions import DAndSError
from musurgia.musurgia_types import check_type


class ArithmeticProgression:
    def __init__(
        self,
        a1: Optional[Union[int, float, Fraction]] = None,
        an: Optional[Union[int, float, Fraction]] = None,
        n: Optional[int] = None,
        d: Optional[Union[int, float, Fraction]] = None,
        s: Optional[Union[int, float, Fraction]] = None,
        correct_s: bool = False,
    ):
        self._a1: Optional[Union[int, float, Fraction]] = None
        self._an: Optional[Union[int, float, Fraction]] = None
        self._n: Optional[Union[int, float, Fraction]] = None
        self._d: Optional[Union[int, float, Fraction]] = None
        self._s: Optional[Union[int, float, Fraction]] = None
        self._current: Fraction
        self._index: int
        self._correction_factor: Optional[Fraction] = None
        self._correct_s: bool

        self.a1 = a1  # type: ignore
        self.an = an  # type: ignore
        self.n = n  # type: ignore
        self.d = d  # type: ignore
        self.s = s  # type: ignore
        self.correct_s = correct_s

    # private methods

    def _check_args(self, arg: Optional[str] = None) -> None:
        if arg is None:
            err = "Not enough attributes are set. Three are needed!"
            if (
                len(
                    [
                        v
                        for v in self._get_private_parameters_dict().values()
                        if v is not None
                    ]
                )
                < 3
            ):
                raise AttributeError(err)
        else:
            if (
                self._get_private_parameters_dict()[arg] is None
                and len(
                    [
                        v
                        for v in self._get_private_parameters_dict().values()
                        if v is not None
                    ]
                )
                > 2
            ):
                err = (
                    "attribute cannot be set. Three parameters are already set. ArithmeticProgression is already "
                    "created!"
                )
                raise AttributeError(err)

    def _calculate_a1(self) -> Fraction:
        if self._d is None:
            return Fraction(2 * self.s, self.n) - self.an
        else:
            return self.an - ((self.n - 1) * self.d)

    def _calculate_an(self) -> Fraction:
        if self._s is None:
            return self.a1 + (self.n - 1) * self.d
        else:
            return Fraction(2 * self.s, self.n) - self.a1

    def _calculate_n(self) -> int:
        if self._s is None:
            output = Fraction((self.an - self.a1), self.d) + 1
        else:
            output = 2 * Fraction(self.s, (self.a1 + self.an))
        return int(output)

    def _calculate_d(self) -> Fraction:
        if self.n == 1:
            output = Fraction(0)
        elif self._a1 is None:
            self._a1 = self._calculate_a1()
            output = Fraction((self.an - self.a1), (self.n - 1))
        elif self._an is None:
            output = Fraction(
                ((self.s - (self.n * self.a1)) * 2), ((self.n - 1) * self.n)
            )
        # elif self._n is None:
        #     self._n = self._calculate_n()
        #     output = Fraction((self.an - self.a1), (self.n - 1))
        else:
            output = Fraction((self.an - self.a1), (self.n - 1))
        return output

    def _calculate_s(self) -> Fraction:
        if self._a1 is None:
            self._a1 = self._calculate_a1()
            output = (self.a1 + self.an) * Fraction(self.n, 2)
        elif self._an is None:
            output = self.n * self.a1 + ((self.n - 1) * Fraction(self.n, 2)) * self.d
        elif self._n is None:
            self._n = self._calculate_n()
            output = (self.a1 + self.an) * Fraction(self.n, 2)
        else:
            output = (self.a1 + self.an) * Fraction(self.n, 2)
        return output

    def _get_private_parameters_dict(
        self,
    ) -> dict[str, Optional[Union[int, Fraction, float]]]:
        return {
            "a1": self._a1,
            "an": self._an,
            "n": self._n,
            "d": self._d,
            "s": self._s,
        }

    def _to_fraction(self, value: Union[int, float, Fraction]) -> Fraction:
        if not isinstance(value, Fraction):
            value = Fraction(value)
        return value

    # public properties

    @property
    def a1(self) -> Fraction:
        """
        >>> arith = ArithmeticProgression(n=3, an=15, d=4)
        >>> arith.a1
        Fraction(7, 1)
        >>> list(arith)
        [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]


        >>> arith = ArithmeticProgression(n=3, an=15, s=33)
        >>> arith.a1
        Fraction(7, 1)
        >>> list(arith)
        [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]


        :return:
        """
        if self._a1 is None:
            self._a1 = self._calculate_a1()
        else:
            self._a1 = self._to_fraction(self._a1)
        return self._a1

    @a1.setter
    def a1(self, value: Optional[Union[int, float, Fraction]]) -> None:
        if value is not None:
            self._check_args("a1")
        self._a1 = value

    @property
    def an(self) -> Fraction:
        """
        >>> arith = ArithmeticProgression(n=3, a1=7, d=4)
        >>> arith.an
        Fraction(15, 1)
        >>> list(arith)
        [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        >>> arith = ArithmeticProgression(n=3, a1=7, s=33)
        >>> arith.an
        Fraction(15, 1)
        >>> list(arith)
        [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        :return:
        """
        if self._an is None:
            self._an = self._calculate_an()
        else:
            self._an = self._to_fraction(self._an)
        return self._an

    @an.setter
    def an(self, value: Optional[Union[int, float, Fraction]]) -> None:
        if value is not None:
            self._check_args("an")
        self._an = value

    @property
    def correct_s(self) -> bool:
        """
        >>> arith = ArithmeticProgression(a1= 3, an=6, s=21)
        >>> arith.get_parameters_dict()
        {'a1': Fraction(3, 1), 'an': Fraction(6, 1), 'n': 4, 'd': Fraction(1, 1), 's': Fraction(21, 1)}
        >>> arith.get_actual_s()
        Fraction(18, 1)
        >>> result = list(arith)
        >>> result
        [Fraction(3, 1), Fraction(4, 1), Fraction(5, 1), Fraction(6, 1)]
        >>> sum(result)
        Fraction(18, 1)

        >>> arith.correct_s = True
        >>> arith.reset_iterator()
        >>> arith.get_correction_factor()
        Fraction(7, 6)
        >>> result = list(arith)
        >>> result
        [Fraction(7, 2), Fraction(14, 3), Fraction(35, 6), Fraction(7, 1)]
        >>> sum(result)
        Fraction(21, 1)

        """
        return self._correct_s

    @correct_s.setter
    def correct_s(self, val: bool) -> None:
        check_type(
            val, bool, class_name=self.__class__.__name__, property_name="correct_s"
        )
        self._correct_s = val
        self._correction_factor = None

    @property
    def d(self) -> Fraction:
        """ "
        >>> arith = ArithmeticProgression(a1= 7, an=15, s=33)
        >>> arith.d
        Fraction(4, 1)
        >>> list(arith)
        [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        >>> arith = ArithmeticProgression(a1= 7, an=15, n=3)
        >>> arith.d
        Fraction(4, 1)
        >>> list(arith)
        [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        >>> arith = ArithmeticProgression(a1= 7, n=3, s=33)
        >>> arith.d
        Fraction(4, 1)
        >>> list(arith)
        [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        >>> arith = ArithmeticProgression(an= 15, n=3, s=33)
        >>> arith.d
        Fraction(4, 1)
        >>> list(arith)
        [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]
        """
        if self._d is None:
            self._d = self._calculate_d()
        else:
            self._d = self._to_fraction(self._d)
        return self._d

    @d.setter
    def d(self, value: Optional[Union[int, float, Fraction]]) -> None:
        if value is not None:
            self._check_args("d")
            if self._s is not None:
                raise DAndSError()
        self._d = value

    @property
    def n(self) -> int:
        """
        >>> arith = ArithmeticProgression(an=15, a1=7, d=4)
        >>> arith.n
        3
        >>> list(arith)
        [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        >>> arith = ArithmeticProgression(an=15, a1=7, s=33)
        >>> arith.n
        3
        >>> list(arith)
        [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        :return:
        """

        if self._n is None:
            self._n = self._calculate_n()
        return int(self._n)

    @n.setter
    def n(self, value: Optional[int]) -> None:
        if value is not None:
            if not isinstance(value, int):
                raise AttributeError("n {} must be int".format(value))
            self._check_args("n")
        self._n = value

    @property
    def s(self) -> Fraction:
        """
        >>> arith = ArithmeticProgression(a1= 7, an=15, d=4)
        >>> arith.s
        Fraction(33, 1)
        >>> list(arith)
        [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        >>> arith = ArithmeticProgression(a1= 7, an=15, n=3)
        >>> arith.s
        Fraction(33, 1)
        >>> list(arith)
        [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        >>> arith = ArithmeticProgression(a1= 7, n=3, d=4)
        >>> arith.s
        Fraction(33, 1)
        >>> list(arith)
        [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        >>> arith = ArithmeticProgression(an= 15, n=3, d=4)
        >>> arith.s
        Fraction(33, 1)
        >>> list(arith)
        [Fraction(7, 1), Fraction(11, 1), Fraction(15, 1)]

        """

        if self._s is None:
            self._s = self._calculate_s()
        else:
            self._s = self._to_fraction(self._s)
        return self._s

    @s.setter
    def s(self, value: Optional[Union[int, float, Fraction]]) -> None:
        if value is not None:
            value = self._to_fraction(value)
            self._check_args("s")
            if self._d is not None:
                raise DAndSError()
        self._s = value

    # public methods

    def get_actual_s(self) -> Fraction:
        return self.n * Fraction((self.a1 + self.an), 2)

    def get_correction_factor(self) -> Fraction:
        def _calculate_correction_factor() -> Fraction:
            if self.correct_s:
                return Fraction(self.s, self.get_actual_s())
            else:
                return Fraction(1, 1)

        if self._correction_factor is None:
            self._correction_factor = _calculate_correction_factor()
        return self._correction_factor

    def get_current_index(self) -> Optional[int]:
        """
        >>> arith = ArithmeticProgression(a1=1, d=2, n=3)
        >>> next(arith)
        Fraction(1, 1)
        >>> arith.get_current_index()
        0
        >>> next(arith)
        Fraction(3, 1)
        >>> arith.get_current_index()
        1
        >>> next(arith)
        Fraction(5, 1)
        >>> arith.get_current_index()
        2
        >>> next(arith)
        Traceback (most recent call last):
            ...
        StopIteration
        >>> arith.get_current_index()
        2
        """
        return self._index

    def get_parameters_dict(self) -> dict[str, Union[int, Fraction]]:
        """
        >>> ArithmeticProgression(n=15, a1=1, d=2).get_parameters_dict()
        {'a1': Fraction(1, 1), 'an': Fraction(29, 1), 'n': 15, 'd': Fraction(2, 1), 's': Fraction(225, 1)}

        :return:
        """
        return {"a1": self.a1, "an": self.an, "n": self.n, "d": self.d, "s": self.s}

    def reset_iterator(self) -> None:
        del self._current
        del self._index

    def __iter__(self) -> "ArithmeticProgression":
        return self

    def __next__(self) -> Fraction:
        try:
            self._current
        except AttributeError:
            self._check_args()
            self._current = self.a1
        try:
            if self._index + 1 >= self.n:
                raise StopIteration()
            self._index += 1
            self._current += self.d
        except AttributeError:
            self._index = 0

        return self._current * self.get_correction_factor()



from unittest import TestCase

from musurgia.trees.fractaltimelinetree import FractalTimelineTree
from musurgia.trees.timelinetree import TimelineDuration


class TestFtSplit(TestCase):
    def setUp(self):
        self.ft = FractalTimelineTree(
            duration=TimelineDuration(10),
            proportions=(1, 2, 3),
            main_permutation_order=(3, 1, 2),
            permutation_index=(1, 1),
        )
        self.ft.add_layer()
        self.first_child = self.ft.get_children()[0]

    def test_two(self):
        s = self.first_child.split(1, 1)
        actual = [x.get_value() for x in s]
        expected = [self.first_child.get_value() / 2, self.first_child.get_value() / 2]
        self.assertEqual(expected, actual)

    def test_three(self):
        s = self.first_child.split(1, 1, 1)
        actual = [x.get_value() for x in s]
        expected = [
            self.first_child.get_value() / 3,
            self.first_child.get_value() / 3,
            self.first_child.get_value() / 3,
        ]
        self.assertEqual(expected, actual)

    def test_child_is_not_leaf(self):
        self.first_child.split(1, 1, 1)
        self.assertFalse(self.first_child.is_leaf)

    def test_split_is_leaf(self):
        s = self.first_child.split(1, 1, 1)
        actual = [x.is_leaf for x in s]
        expected = [True, True, True]
        self.assertEqual(expected, actual)



import copy
from pathlib import Path
from unittest import TestCase

from musurgia.pdf.line import HorizontalLineSegment, StraightLine
from musurgia.pdf.pdf import Pdf
from musurgia.tests.utils_for_tests import PdfTestCase

path = Path(__file__)


class TestHorizontalLineSegment(TestCase):
    def setUp(self) -> None:
        self.hls = HorizontalLineSegment(length=10)

    def test_relative_y(self):
        assert self.hls.start_mark_line.relative_y == 0
        assert self.hls.end_mark_line.relative_y == 0

        self.hls.start_mark_line.length = 10
        self.hls.end_mark_line.length = 6

        assert self.hls.relative_y == 0
        assert self.hls.start_mark_line.relative_y == 0
        assert self.hls.end_mark_line.relative_y == 2
        assert self.hls.straight_line.relative_y == 5

        assert self.hls.get_relative_y2() == 10

        assert self.hls.get_height() == 10
        assert self.hls.get_relative_y2() == 10

        self.hls.relative_y = 10
        assert self.hls.relative_y == 10
        assert self.hls.start_mark_line.relative_y == 10
        assert self.hls.end_mark_line.relative_y == 12
        assert self.hls.straight_line.relative_y == 15
        assert self.hls.get_height() == 10
        assert self.hls.get_relative_y2() == 20

        self.hls.set_straight_line_relative_y(0)
        assert self.hls.relative_y == -5
        assert self.hls.start_mark_line.relative_y == -5
        assert self.hls.end_mark_line.relative_y == -3
        assert self.hls.straight_line.relative_y == 0
        assert self.hls.get_height() == 10
        assert self.hls.get_relative_y2() == 5

        self.hls.start_mark_line.length = 6
        self.hls.end_mark_line.length = 10
        assert self.hls.start_mark_line.relative_y == -3
        assert self.hls.end_mark_line.relative_y == -5

        self.hls.relative_y = 0
        assert self.hls.start_mark_line.relative_y == 2
        assert self.hls.end_mark_line.relative_y == 0

    def test_relative_x(self):
        self.hls.relative_x = 10
        self.hls.right_margin = 5

        assert (
            self.hls.straight_line.relative_x
            == self.hls.start_mark_line.relative_x
            == self.hls.relative_x
            == 10
        )
        assert self.hls.end_mark_line.relative_x == 10 + self.hls.length

        assert self.hls.get_relative_x2() == 20
        assert self.hls.get_width() == 15

    def test_straight_line_top_margin(self):
        actual = self.hls.straight_line.top_margin
        expected = 0
        self.assertEqual(expected, actual)

    def test_start_mark_line_top_margin(self):
        actual = self.hls.start_mark_line.top_margin
        expected = 0
        self.assertEqual(expected, actual)

    def test_set_and_get_length(self):
        assert self.hls.length == 10
        self.hls.length = 20
        assert self.hls.length == 20

    def test_infos(self):
        hls = HorizontalLineSegment(length=100)
        hls.start_mark_line.length = 6
        hls.end_mark_line.length = 10
        hls.relative_x = 5
        hls.relative_y = 5
        assert hls.get_positions() == {"x": 5.0, "y": 5.0}
        assert (hls.get_relative_x2(), hls.get_relative_y2()) == (105.0, 15.0)
        assert hls.get_height() == 10
        assert hls.get_width() == 100

    def test_set_straight_line_y(self):
        self.hls.start_mark_line.length = self.hls.end_mark_line.length = 5
        self.hls.relative_y = 10
        assert self.hls.straight_line.relative_y == 12.5
        self.hls.set_straight_line_relative_y(0)
        assert self.hls.relative_y == -2.5
        self.hls.set_straight_line_relative_y(5)
        assert self.hls.relative_y == 2.5
        self.hls.relative_y = 0
        self.hls.start_mark_line.length = 10
        assert self.hls.straight_line.relative_y == 5
        self.hls.end_mark_line.length = 15
        assert self.hls.straight_line.relative_y == 7.5
        self.hls.set_straight_line_relative_y(0)
        assert self.hls.straight_line.relative_y == 0


class TestTestHorizontalLineSegment(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.hls = HorizontalLineSegment(length=10)

    def test_draw(self):
        with self.file_path(parent_path=path, name="draw", extension="pdf") as pdf_path:
            self.pdf.translate_page_margins()
            self.hls.end_mark_line.show = True
            self.hls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_draw_list(self):
        segments = [HorizontalLineSegment(length) for length in range(10, 30, 5)]
        segments[-1].end_mark_line.show = True
        with self.file_path(
            parent_path=path, name="draw_list", extension="pdf"
        ) as pdf_path:
            self.pdf.translate_page_margins()
            for segment in segments:
                segment.draw(self.pdf)
                self.pdf.translate(segment.get_width(), 0)

            self.pdf.write_to_path(pdf_path)

    def test_labels(self):
        self.hls.relative_x = 30
        self.hls.length = 30
        self.hls.start_mark_line.length = 10
        self.hls.end_mark_line.length = 10
        self.hls.end_mark_line.add_label("end top")
        self.hls.end_mark_line.add_label("end top")
        self.hls.start_mark_line.add_label("start top")
        self.hls.start_mark_line.add_label("start top")
        self.hls.end_mark_line.add_label("end below", placement="below")
        self.hls.end_mark_line.add_label("end below", placement="below")
        self.hls.start_mark_line.add_label("start below", placement="below")
        self.hls.start_mark_line.add_label("start below", placement="below")
        self.hls.straight_line.add_label("straight above", placement="above")
        self.hls.straight_line.add_label("straight below", placement="below")
        self.hls.straight_line.add_label("straight left", placement="left")
        self.hls.straight_line.add_label("straight left", placement="left")
        self.hls.straight_line.add_label("straight left", placement="left")

        with self.file_path(
            parent_path=path, name="labels", extension="pdf"
        ) as pdf_path:
            self.pdf.translate_page_margins()
            self.hls.end_mark_line.show = True
            self.hls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_draw_with_relative_y(self):
        copied_1 = copy.deepcopy(self.hls)
        copied_1.relative_y = 5
        control = StraightLine(mode="h", length=10, relative_y=5)
        control.add_text_label("control", font_size=8, bottom_margin=2)

        with self.file_path(
            parent_path=path, name="draw_with_relative_y", extension="pdf"
        ) as pdf_path:
            self.pdf.translate_page_margins()
            # self.hls.start_mark_line.add_text_label(str(self.pdf.absolute_positions), placement='above', font_size=8)
            self.hls.draw(self.pdf)
            self.pdf.translate(20, 0)
            control.draw(self.pdf)
            # copied_1.start_mark_line.add_text_label(str(self.pdf.absolute_positions), placement='above', font_size=8)
            copied_1.start_mark_line.add_text_label(
                f"hls: {copied_1.positions}", placement="below", font_size=8
            )
            copied_1.start_mark_line.add_text_label(
                f"straight: {copied_1.straight_line.positions}",
                placement="below",
                font_size=8,
            )
            copied_1.start_mark_line.add_text_label(
                f"start: {copied_1.start_mark_line.positions}",
                placement="below",
                font_size=8,
            )
            copied_1.start_mark_line.add_text_label(
                f"end: {copied_1.end_mark_line.positions}",
                placement="below",
                font_size=8,
            )

            copied_1.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)



from abc import abstractmethod
from copy import deepcopy
from enum import Enum
from itertools import cycle
from typing import Any, Iterator, Literal, Optional, Union

from musicscore.midi import Midi
from musicscore.score import Score
from musurgia.chordfactory.chordfactory import AbstractChordFactory
from musurgia.trees.fractaltimelinetree import FractalTimelineTree
from musurgia.trees.timelinetree import TimelineTree
from musurgia.utils import RelativeValueGenerator


class TreeChordFactory(AbstractChordFactory):
    def __init__(
        self,
        musical_tree_node: "MusicalTree",
        show_metronome: bool = False,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._musical_tree_node: "MusicalTree" = musical_tree_node
        self._show_metronome: bool
        self.show_metronome = show_metronome

    @property
    def show_metronome(self) -> bool:
        return self._show_metronome

    @show_metronome.setter
    def show_metronome(self, value: bool) -> None:
        self._show_metronome = value

    def get_musical_tree_node(self) -> "MusicalTree":
        return self._musical_tree_node

    def update_chord_quarter_duration(self) -> None:
        self._chord.quarter_duration = deepcopy(
            self.get_musical_tree_node().get_duration().get_quarter_duration()
        )

    @abstractmethod
    def update_chord_midis(self) -> None:
        pass

    def update_chord_metronome(self) -> None:
        if self.show_metronome:
            self._chord.metronome = (
                self.get_musical_tree_node().get_duration().get_metronome()
            )
        else:
            self._chord._metronome = None


class MusicalTree(TimelineTree):
    @abstractmethod
    def get_chord_factory(self) -> TreeChordFactory:
        pass

    def export_score(self) -> Score:
        score = Score()
        for layer_number in range(self.get_number_of_layers() + 1):
            part = score.add_part(f"part-{layer_number + 1}")
            layer = self.get_layer(level=layer_number)
            for node in layer:
                part.add_chord(node.get_chord_factory().create_chord())
        return score


class MidiMusicalTreeChordFactory(TreeChordFactory):
    def __init__(
        self, midis: list[Midi] = [Midi(72)], *args: Any, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self._midis: list[Midi]
        self.midis = midis

    def update_chord_midis(self) -> None:
        self._chord.midis = deepcopy(self.get_midis())

    @property
    def midis(self) -> list[Midi]:
        return self._midis

    @midis.setter
    def midis(self, value: list[Midi]) -> None:
        self._midis = value

    def get_midis(self) -> list[Midi]:
        return self.midis


class MidiMusicalTree(MusicalTree):
    def __init__(
        self, midis: list[Midi] = [Midi(72)], *args: Any, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self._tree_chord_factory: MidiMusicalTreeChordFactory = (
            MidiMusicalTreeChordFactory(musical_tree_node=self, midis=midis)
        )

    def get_chord_factory(self) -> TreeChordFactory:
        return self._tree_chord_factory


class FractalMusicalTree(FractalTimelineTree, MidiMusicalTree):
    pass


MidiValue = Union[int, float]


class MidiValueMicroTone(Enum):
    HALF = 1.0
    QUARTER = 0.5
    EIGHT = 0.25

DirectionValue = Literal[-1, 1]

class RelativeMusicTree(MusicalTree):
    TreeChordFactoryClass = MidiMusicalTreeChordFactory

    def __init__(
        self,
        midi_value_range: Optional[tuple[MidiValue]] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._midi_value_range: Optional[tuple[MidiValue]] = midi_value_range
        self._micro_tone: MidiValueMicroTone = MidiValueMicroTone.HALF
        self._direction_iterator: Iterator[DirectionValue] = cycle([-1, 1])
        self._tree_chord_factory = (
            self.TreeChordFactoryClass(musical_tree_node=self)
        )

    @property
    def micro_tone(self) -> MidiValueMicroTone:
        return self._micro_tone

    @micro_tone.setter
    def micro_tone(self, value: MidiValueMicroTone) -> None:
        self._micro_tone = value

    @property
    def midi_value_range(self) -> Optional[tuple[MidiValue]]:
        return self._midi_value_range

    @midi_value_range.setter
    def midi_value_range(self, value: Optional[tuple[MidiValue]]) -> None:
        self._midi_value_range = value

    @property
    def direction_iterator(self) -> Iterator[DirectionValue]:
        return self._direction_iterator

    @direction_iterator.setter
    def direction_iterator(self, value: Iterator[DirectionValue]) -> None:
        self._direction_iterator = value

    def get_chord_factory(self) -> TreeChordFactory:
        return self._tree_chord_factory

    def set_relative_midis(self) -> None:
        if not self.midi_value_range:
            raise AttributeError("set_relative_midis: midi_value_range cannot be None.")
        for node in self.traverse():
            node.get_chord_factory().midis = [Midi(node.midi_value_range[0])]
            if not node.is_leaf:
                children = node.get_children()
                proportions = [ch.get_value() for ch in children]
                directions = [
                    next(node.direction_iterator) for _ in range(len(proportions))
                ]
                children_midi_value_ranges = list(
                    RelativeValueGenerator(
                        value_range=node.midi_value_range,
                        directions=directions,
                        proportions=proportions,
                        value_grid=self.micro_tone.value,
                    )
                )
                for index in range(len(children_midi_value_ranges) - 1):
                    min_midi = float(children_midi_value_ranges[index])
                    max_midi = float(children_midi_value_ranges[index + 1])
                    children[index].midi_value_range = (min_midi, max_midi)


class FractalDirectionIterator:
    def __init__(self, main_direction_cell: list[DirectionValue], fractal_node: FractalTimelineTree, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._fractal_node: FractalTimelineTree = fractal_node
        self._main_direction_cell: list[DirectionValue] = main_direction_cell
        self._iter_index = -1

    def reset(self) -> None:
        self._iter_index = -1

    def get_directions(self) -> list[DirectionValue]:
        fractal_orders = [
            ch.get_fractal_order() for ch in self._fractal_node.get_children()
        ]
        return [self.get_main_directions()[fo - 1] for fo in fractal_orders]

    def get_main_directions(self) -> list[DirectionValue]:
        cy = cycle(self._main_direction_cell)
        return [next(cy) for _ in range(len(self._fractal_node.proportions))]

    def __iter__(self) -> Iterator[DirectionValue]:
        return self

    def __next__(self) -> DirectionValue:
        try:
            self._iter_index += 1
            return self.get_directions()[self._iter_index]
        except IndexError:
            raise StopIteration


class FractalRelativeMusicTree(FractalTimelineTree, RelativeMusicTree):
    def __init__(self, main_direction_cell: list[DirectionValue]=[1, -1], *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._direction_iterator = FractalDirectionIterator(
            main_direction_cell=main_direction_cell, fractal_node=self
        )



import copy
from pathlib import Path
from unittest import TestCase

from musurgia.musurgia_exceptions import DrawObjectInContainerHasNegativePositionError
from musurgia.pdf.line import (
    HorizontalLineSegment,
    VerticalSegmentedLine,
    VerticalLineSegment,
    StraightLine,
    HorizontalSegmentedLine,
)
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.pdf_tools import draw_ruler
from musurgia.pdf.rowcolumn import DrawObjectRow, DrawObjectColumn
from musurgia.pdf.labeled import TextLabel
from musurgia.tests.utils_for_tests import (
    PdfTestCase,
    add_control_positions_to_draw_object,
    create_simple_column,
)

path = Path(__file__)


def _make_draw_objects():
    h_segments = [HorizontalLineSegment(10), HorizontalLineSegment(20)]
    h_segments[1].start_mark_line.length = 6

    vsl = VerticalSegmentedLine(lengths=[5, 6, 7, 8])
    return h_segments, vsl


def make_row():
    r = DrawObjectRow()
    h_segments, vsl = _make_draw_objects()
    for segment in h_segments:
        r.add_draw_object(segment)
    r.add_draw_object(vsl)
    r.add_text_label(TextLabel("row", placement="left"))
    return r


def make_column():
    c = DrawObjectColumn()
    h_segments, vsl = _make_draw_objects()
    for segment in h_segments:
        c.add_draw_object(segment)
    c.add_draw_object(vsl)
    return c


class TestRowPositionAndMargins(TestCase):
    def setUp(self):
        self.r = DrawObjectRow()
        self.hls = HorizontalLineSegment(length=10)
        self.hls.start_mark_line.length = 5
        self.r.add_draw_object(self.hls)

    def test_default_positions(self):
        assert self.r.positions == self.hls.positions == (0, 0)
        assert self.r.get_relative_x2() == self.hls.get_relative_x2() == 10
        assert self.r.get_relative_y2() == self.hls.get_relative_y2() == 5

    def test_change_row_positions(self):
        self.r.positions = (20, 30)
        assert self.hls.positions == (0, 0)
        assert self.r.get_relative_x2() == 30
        assert self.r.get_relative_y2() == 35
        assert self.hls.positions == (0, 0)
        assert self.hls.get_relative_x2() == 10
        assert self.hls.get_relative_y2() == 5

    def test_change_hls_positions(self):
        self.hls.positions = (20, 30)
        assert self.hls.get_relative_x2() == 30
        assert self.hls.get_relative_y2() == 35
        assert self.hls.get_width() == 10

        assert self.r.positions == (0, 0)
        assert self.r.get_relative_x2() == 30
        assert self.r.get_relative_y2() == 35

    def test_change_hls_negative_positions_error(self):
        self.hls.positions = (-20, -30)
        assert self.hls.get_relative_x2() == -10
        assert self.hls.get_relative_y2() == -25

        with self.assertRaises(DrawObjectInContainerHasNegativePositionError):
            self.r.draw(Pdf())

    def test_change_hls_and_row_positions(self):
        self.r.positions = (20, 30)
        self.hls.positions = (40, 50)
        assert self.hls.length == 10
        assert self.hls.get_end_positions() == (50, 55)
        assert self.r.get_end_positions() == (70, 85)

    def test_change_one_hls_positions(self):
        hl2 = copy.deepcopy(self.hls)
        self.r.add_draw_object(hl2)
        hl2.positions = (20, 30)
        assert self.hls.positions == (0, 0)
        assert self.hls.get_relative_x2(), self.hls.get_relative_y2() == (10, 5)
        assert hl2.positions == (20, 30)
        assert (hl2.get_relative_x2(), hl2.get_relative_y2()) == (30, 35)

        assert self.r.get_relative_x2() == 40
        assert self.r.get_relative_y2() == 35

    def test_relative_y_inside_column(self):
        c = DrawObjectColumn()
        hsl = HorizontalSegmentedLine([10, 20, 30])
        c.add_draw_object(hsl)
        assert c.get_end_positions() == (60, 3)
        hsl.relative_y = 10
        hsl.margins = (5, 0, 15, 0)
        assert hsl.get_height() == 23
        assert hsl.get_relative_y2() == 13
        assert c.get_end_positions() == (60, 33)

    def test_two_simple_lines_column_draw(self):
        c = DrawObjectColumn()
        s1 = StraightLine("h", 30)
        s1.relative_y = 5
        s1.margins = (10, 10, 10, 10)
        s2 = StraightLine("h", 30)
        s2.margins = (10, 10, 10, 10)
        s2.relative_y = 5
        c.add_draw_object(s1)
        c.add_draw_object(s2)
        assert c.get_end_positions() == (50, 50)

    def test_get_border_rectangle_coordinates(self):
        assert self.r.get_border_rectangle_coordinates() == (0, 0, 10, 5)

    def test_get_margin_rectangle_coordinates(self):
        assert self.r.get_margin_rectangle_coordinates() == (0, 0, 10, 5)


class TestRowColumnSimpleLines(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf(orientation="l")
        self.h_lines = [StraightLine(length=l, mode="h") for l in range(5, 20, 5)]
        self.v_lines = [StraightLine(length=l, mode="v") for l in range(5, 20, 5)]
        self.v_lines[-1].right_margin = 10

    def test_simple_line_row_positions(self):
        row = DrawObjectRow(show_margins=True, show_borders=True)
        for l in self.h_lines + self.v_lines:
            l.top_margin = 10
            l.left_margin = 10
            row.add_draw_object(l)

        assert self.v_lines[2].get_end_positions() == (0.0, 15.0)
        assert self.v_lines[2].get_height() == 25
        add_control_positions_to_draw_object(row)
        assert row.get_end_positions() == (100, 25)
        with self.file_path(path, "simple_lines_row_positions", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(mode="v", pdf=self.pdf)
            draw_ruler(mode="h", pdf=self.pdf)
            self.pdf.translate(10, 10)
            row.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_simple_lines_row(self):
        row = DrawObjectRow(show_margins=True, show_borders=True)
        row.margins = (10, 10, 10, 10)
        end_line = StraightLine(mode="h", length=70)
        end_line.add_text_label("control end_line", bottom_margin=2)
        for l in self.h_lines + self.v_lines:
            row.add_draw_object(l)

        for l in self.h_lines + self.v_lines:
            copied = copy.deepcopy(l)
            copied.top_margin = 10
            copied.left_margin = 10
            row.add_draw_object(copied)

        add_control_positions_to_draw_object(row)
        row.add_text_label(f"{row.get_height()}", placement="left")
        with self.file_path(path, "simple_lines_row", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(mode="v", pdf=self.pdf)
            draw_ruler(mode="h", pdf=self.pdf)
            self.pdf.translate(20, 20)
            row.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)


class TestRowColumn(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf(orientation="l")
        self.row = make_row()
        self.column = make_column()

    def test_draw_row(self):
        self.pdf.translate_page_margins()
        draw_ruler(self.pdf, "h")
        draw_ruler(self.pdf, "v")
        self.pdf.translate(10, 10)
        r = self.row
        r.add_text_label(label=TextLabel("below label", placement="below"))

        r.draw(self.pdf)

        with self.file_path(path, "row", "pdf") as pdf_path:
            self.pdf.write_to_path(pdf_path)

    def test_wrong_add_draw_object(self):
        r = DrawObjectRow()
        with self.assertRaises(TypeError):
            r.add_draw_object("something")

    def test_draw_row_of_segments(self):
        draw_object_rows = [
            DrawObjectRow(show_segments=True, show_margins=True) for _ in range(4)
        ]
        line_segments = [HorizontalLineSegment(l) for l in [30, 10, 20]]
        line_segments[-1].end_mark_line.show = True
        line_segments[-1].right_margin = 10
        for l in line_segments:
            draw_object_rows[0].add_draw_object(l)

        for i, l in enumerate(line_segments):
            copied = copy.deepcopy(l)
            copied.start_mark_line.length *= i + 1
            draw_object_rows[1].add_draw_object(copied)
            if i == len(line_segments) - 1:
                copied.end_mark_line.length += copied.start_mark_line.length
            copied.right_margin = 10

        for i, l in enumerate(line_segments):
            copied = copy.deepcopy(l)
            copied.start_mark_line.length *= i + 1
            draw_object_rows[2].add_draw_object(copied)
            if i == len(line_segments) - 1:
                copied.end_mark_line.length += copied.start_mark_line.length

        for i, l in enumerate(line_segments):
            copied = copy.deepcopy(l)
            copied.start_mark_line.length *= i + 1
            draw_object_rows[3].add_draw_object(copied)
            if i == len(line_segments) - 1:
                copied.end_mark_line.length += copied.start_mark_line.length

        with self.file_path(path, "row_of_segments", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            for l in line_segments:
                copied = copy.deepcopy(l)
                copied.draw(self.pdf)
                self.pdf.translate(50, 0)

            self.pdf.reset_position()
            self.pdf.translate_page_margins()
            self.pdf.translate(0, 30)
            for i, l in enumerate(line_segments):
                copied = copy.deepcopy(l)
                copied.start_mark_line.length *= i + 1
                copied.draw(self.pdf)
                self.pdf.translate(50, 0)

            self.pdf.reset_position()
            self.pdf.translate_page_margins()
            self.pdf.translate(0, 60)
            for i, l in enumerate(line_segments):
                copied = copy.deepcopy(l)
                copied.start_mark_line.length *= i + 1
                copied.relative_y -= i * 3
                copied.start_mark_line.add_text_label(
                    str(copied.relative_y), font_size=8
                )
                copied.start_mark_line.add_text_label(
                    str(copied.start_mark_line.length), font_size=8, placement="left"
                )
                copied.straight_line.add_text_label(
                    str(copied.straight_line.length),
                    font_size=8,
                    placement="below",
                    top_margin=2,
                )
                copied.draw(self.pdf)
                self.pdf.translate(50, 0)

            self.pdf.reset_position()
            self.pdf.translate_page_margins()
            self.pdf.translate(0, 90)
            for do in draw_object_rows:
                do.draw(self.pdf)
                self.pdf.translate(0, 20)

            self.pdf.write_to_path(pdf_path)

    def test_draw_column_of_row_of_segments(self):
        r = DrawObjectRow()
        r.add_draw_object(HorizontalLineSegment(30))
        r.add_draw_object(HorizontalLineSegment(10))
        r.add_draw_object(HorizontalLineSegment(20))
        c = DrawObjectColumn()
        c.add_draw_object(HorizontalLineSegment(60))
        c.add_draw_object(r)

        with self.file_path(path, "column_of_row_of_segments", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(10, 10)
            c.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_draw_row_of_column_of_segments(self):
        c = DrawObjectColumn()
        c.add_draw_object(HorizontalLineSegment(30))
        c.add_draw_object(HorizontalLineSegment(10))
        c.add_draw_object(HorizontalLineSegment(20))
        r = DrawObjectRow()
        r.add_draw_object(c)

        with self.file_path(path, "row_of_column_of_segments", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(10, 10)
            c.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_draw_column(self):
        self.pdf.translate_page_margins()
        draw_ruler(self.pdf, "h")
        draw_ruler(self.pdf, "v")
        self.pdf.translate(10, 10)
        c = self.column
        c.add_text_label(TextLabel("below label", placement="below", top_margin=3))
        c.draw(self.pdf)
        with self.file_path(path, "column", "pdf") as pdf_path:
            self.pdf.write_to_path(pdf_path)

    def test_travers_column(self):
        c = DrawObjectColumn()
        r = DrawObjectRow()
        c.add_draw_object(r)
        c_1 = c.add_draw_object(HorizontalLineSegment(length=10))
        r_1 = r.add_draw_object(HorizontalLineSegment(length=20))
        assert list(c.traverse()) == [c, r, r_1, c_1]

    def test_travers_row(self):
        r = DrawObjectRow()
        c = DrawObjectColumn()
        cc = DrawObjectColumn()

        r.add_draw_object(c)
        r_1 = r.add_draw_object(HorizontalLineSegment(length=10))
        r.add_draw_object(cc)
        rr = DrawObjectRow()
        c.add_draw_object(rr)
        rr_1 = rr.add_draw_object(HorizontalLineSegment(length=10))
        c_1 = c.add_draw_object(HorizontalLineSegment(length=10))
        cc_1 = cc.add_draw_object(HorizontalLineSegment(length=10))
        cc_2 = cc.add_draw_object(HorizontalLineSegment(length=10))
        assert list(r.traverse()) == [r, c, rr, rr_1, c_1, r_1, cc, cc_1, cc_2]

    def test_row_column_borders_margins_and_labels(self):
        def add_text_labels(do):
            if isinstance(do, HorizontalLineSegment):
                do = do.start_mark_line
            do.add_text_label(label=TextLabel("below label", placement="below"))
            do.add_text_label(
                label=TextLabel(f"rel_x: {round(do.relative_x)}", placement="below")
            )
            do.add_text_label(
                label=TextLabel(
                    f"rel_x2: {round(do.get_relative_x2())}", placement="below"
                )
            )
            do.add_text_label(label=TextLabel("above label", placement="above"))
            do.add_text_label(
                label=TextLabel(f"rel_y: {round(do.relative_y)}", placement="above")
            )
            do.add_text_label(
                label=TextLabel(
                    f"rel_y2: {round(do.get_relative_y2())}", placement="above"
                )
            )
            do.add_text_label(label=TextLabel("left label", placement="left"))
            do.add_text_label(label=TextLabel("left label", placement="left"))
            do.add_text_label(label=TextLabel("left label", placement="left"))
            for tl in do.get_text_labels():
                tl.font_size = 8

        hls = HorizontalLineSegment(length=20)
        hls.start_mark_line.length = 10

        vls = VerticalLineSegment(length=15)
        vls.start_mark_line.length = vls.end_mark_line.length = 10

        control_hls = copy.deepcopy(hls)
        control_hls.top_margin = 10
        control_hls.left_margin = 20
        control_hls.bottom_margin = 20

        add_text_labels(control_hls)
        control_hls.straight_line.add_text_label("control_hsl")

        r = DrawObjectRow(show_borders=True, show_margins=True)
        r.relative_y = 10
        first_hls = copy.deepcopy(hls)
        first_hls.left_margin = 5
        second_hls = copy.deepcopy(hls)
        second_hls.start_mark_line.length *= 0.5
        second_hls.end_mark_line.show = True

        r.add_draw_object(first_hls)
        r.add_draw_object(second_hls)

        first_vls = copy.deepcopy(vls)
        second_vls = copy.deepcopy(vls)

        second_vls.end_mark_line.show = first_vls.end_mark_line.show = True
        first_vls.left_margin = 10
        first_vls.right_margin = 5
        r.add_draw_object(first_vls)
        r.add_draw_object(second_vls)

        r.top_margin = 10
        r.left_margin = 20
        r.bottom_margin = 20

        add_text_labels(r)

        c = DrawObjectColumn(show_borders=True, show_margins=True)
        first_hls = copy.deepcopy(hls)
        second_hls = copy.deepcopy(hls)
        first_hls.bottom_margin = 10
        c.add_draw_object(first_hls)
        c.add_draw_object(second_hls)
        c.top_margin = 10
        c.left_margin = 20

        add_text_labels(c)

        main_column = DrawObjectColumn()
        main_column.add_draw_object(control_hls)
        main_column.add_draw_object(r)
        main_column.add_draw_object(c)

        with self.file_path(path, "borders_margins_and_labels", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, mode="v")
            draw_ruler(self.pdf, mode="h")
            self.pdf.translate(20, 20)
            main_column.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    # def test_complex_column(self):
    #     c1 = DrawObjectColumn()
    #     hl1 = HorizontalSegmentedLine([10, 20, 30, 40])
    #     c1.add_draw_object(hl1)
    #     first_row = DrawObjectRow()
    #     second_row = DrawObjectRow()
    #     c1.add_draw_object(first_row)
    #     c1.add_draw_object(second_row)
    #
    #     c2 = DrawObjectColumn()
    #     first_row.add_draw_object(c2)
    #     hl2 = HorizontalSegmentedLine([40, 30, 20, 10])
    #     c2.add_draw_object(hl2)
    #     c2.add_draw_object(second_row)
    #
    #     c3 = DrawObjectColumn()
    #     second_row.add_draw_object(c3)
    #     hl3 = HorizontalSegmentedLine([5, 10, 15, 20])
    #     hl4 = HorizontalSegmentedLine([20, 15, 10, 5])
    #     c3.add_draw_object(hl3)
    #     c3.add_draw_object(hl4)
    #
    #     c = DrawObjectColumn()
    #     c.add_draw_object(c1)
    #     hl5 = HorizontalSegmentedLine([50, 40, 30, 20, 10])
    #     c.add_draw_object(hl5)
    #
    #     hl1.bottom_margin = 10
    #     hl2.position = (30, 20)
    #     hl2.position = (10, 20)
    #     hl3.position = (40, 50)
    #     hl4.position = (20, 10)
    #
    #     hl1.add_text_label('hl1', placement='left')
    #     hl2.add_text_label('hl2', placement='left')
    #     hl3.add_text_label('hl3', placement='left')
    #     hl4.add_text_label('hl4', placement='left')
    #     hl5.add_text_label('hl5', placement='left')
    #
    #     with self.file_path(path, 'complex_column', 'pdf') as pdf_path:
    #         self.pdf.translate_page_margins()
    #         draw_ruler(self.pdf, mode='v', first_label=-1)
    #         draw_ruler(self.pdf, mode='h', first_label=-1)
    #         self.pdf.translate(20, 20)
    #         c.draw(self.pdf)
    #         self.pdf.write_to_path(pdf_path)

    def test_two_simple_lines_column_draw(self):
        c = create_simple_column([StraightLine("h", 30), StraightLine("h", 30)])

        with self.file_path(path, "two_simple_lines_column", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, mode="v", first_label=-1)
            draw_ruler(self.pdf, mode="h", first_label=-1)
            self.pdf.translate(10, 10)
            c.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_simple_and_segmented_lines_in_columns(self):
        control1 = HorizontalSegmentedLine([10, 20])
        control1.margins = (10, 10, 10, 10)
        control2 = StraightLine("h", 30)
        control2.margins = (10, 10, 10, 10)

        r = DrawObjectRow()
        c1 = create_simple_column([StraightLine("h", 30), StraightLine("h", 30)])
        c1.right_margin = 10
        c2 = create_simple_column(
            [HorizontalSegmentedLine([10, 20]), HorizontalSegmentedLine([10, 20])]
        )
        for c in [c1, c2]:
            r.add_draw_object(c)

        with self.file_path(
            path, "simple_and_segmented_lines_in_columns", "pdf"
        ) as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, mode="v", first_label=-1)
            draw_ruler(self.pdf, mode="h", first_label=-1)
            self.pdf.translate(10, 10)
            control1.draw(self.pdf)
            self.pdf.translate(50, 0)
            control2.draw(self.pdf)
            self.pdf.translate(50, 0)
            r.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_two_horizontal_segmented_lines_column_draw(self):
        c = create_simple_column(
            [HorizontalSegmentedLine([10, 20]), HorizontalSegmentedLine([10, 20])]
        )

        with self.file_path(
            path, "two_horizontal_segmented_lines_column", "pdf"
        ) as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, mode="v", first_label=-1)
            draw_ruler(self.pdf, mode="h", first_label=-1)
            self.pdf.translate(10, 10)
            c.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_relative_y_inside_column_draw(self):
        c = DrawObjectColumn(show_borders=True, show_margins=True)
        hsl = HorizontalSegmentedLine([10, 20, 30])
        hsl.relative_y = 20
        hsl.margins = (10, 0, 20, 0)
        c.add_draw_object(hsl)

        line = StraightLine(length=60, mode="h")
        c.add_draw_object(line)
        with self.file_path(path, "relative_y_inside_column", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, mode="v", first_label=-1)
            draw_ruler(self.pdf, mode="h", first_label=-1)
            self.pdf.translate(10, 10)
            c.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)



from musurgia.pdf.pdfunit import PdfUnit
from musurgia.tests.utils_for_tests import PdfTestCase


class TestPdfK(PdfTestCase):
    def test_get_k(self):
        actual = PdfUnit.get_k()
        expected = 2.834645669291339
        self.assertEqual(expected, actual)

    def test_new_unit(self):
        PdfUnit.GLOBAL_UNIT = "cm"
        actual = PdfUnit.get_k()
        expected = 28.346456692913385
        self.assertEqual(expected, actual)
        PdfUnit.reset()

    def test_change_unit(self):
        PdfUnit.change("mm")
        old_k = PdfUnit.get_k()
        PdfUnit.change("cm")
        new_k = PdfUnit.get_k()
        actual = new_k / old_k
        expected = 10
        self.assertAlmostEqual(expected, actual)
        PdfUnit.reset()



from pathlib import Path

from musurgia.pdf.line import HorizontalLineSegment
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.pdf_tools import draw_ruler
from musurgia.pdf.text import Text
from musurgia.tests.utils_for_tests import PdfTestCase

path = Path(__file__)


class TestText(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_font_attributes(self):
        t = Text("Test")
        t.font_size = 20
        t.font_weight = "medium"
        t.font_style = "italic"
        t.font_family = "Courier"
        assert t.font_size == 20
        assert t.font_weight == "medium"
        assert t.font_style == "italic"
        assert t.font_family == "Courier"

        with self.assertRaises(TypeError):
            t.font_style = "wrong"

    def test_draw(self):
        t = Text("The fox is going to be dead.")
        with self.file_path(path, "draw", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(10, 10)
            t.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_draw_with_top_margin(self):
        t = Text("The fox is going to be dead.")
        t.top_margin = 5
        with self.file_path(path, "draw_with_top_margin", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(10, 10)
            t.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_get_height(self):
        t = Text("fox is going to be dead.")
        t.font_size = 14
        t.top_margin = 3
        self.assertEqual(7.02, round(t.get_height(), 3))

    def test_height_graphical(self):
        t = Text("The fox is going to be dead.")
        with self.file_path(path, "height_graphical", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(10, 10)
            hls = HorizontalLineSegment(length=t.get_text_width())
            hls.relative_y = t.relative_y - 3 / 4 * t.get_height()
            hls.start_mark_line.length = t.get_height()
            hls.end_mark_line.show = True

            with self.pdf.saved_state():
                self.pdf.rect(
                    hls.relative_x, hls.relative_y, hls.get_width(), t.get_height()
                )
                hls.draw(self.pdf)
            t.draw(self.pdf)

            self.pdf.write_to_path(pdf_path)

    def test_draw_multiple(self):
        t1 = Text(value="Fox is going to be dead.")
        t2 = Text(value="What should we do??", relative_y=0)
        t3 = Text(value="What should we do??", relative_y=0)
        with self.file_path(path, "draw_multiple", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(10, 10)
            t1.draw(self.pdf)
            self.pdf.translate(0, t1.get_height())
            t2.draw(self.pdf)
            self.pdf.translate(0, t1.get_height())
            t3.draw(self.pdf)
            self.pdf.translate(0, t1.get_height())
            self.pdf.write_to_path(pdf_path)

    def test_draw_multiple_with_relative_y(self):
        t1 = Text(value="Fox is going to be dead.", relative_x=10, relative_y=10)
        t2 = Text(value="What should we do??", relative_x=10, relative_y=20)
        t3 = Text(value="What should we do??", relative_x=10, relative_y=30)

        with self.file_path(path, "draw_multiple_with_relative_y", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")

            t1.draw(self.pdf)
            t2.draw(self.pdf)
            t3.draw(self.pdf)

            self.pdf.write_to_path(pdf_path)

    def test_text_font(self):
        t = Text(
            value="some text",
            font_family="Courier",
            font_weight="bold",
            font_style="italic",
        )
        assert t.font_family == "Courier"
        assert t.font_weight == "bold"
        assert t.font_style == "italic"

        t2 = Text(value="some text", font_style="italic")

        with self.file_path(path, "font", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            t.draw(self.pdf)
            self.pdf.translate(dx=0, dy=10)
            t2.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)



from fractions import Fraction
from math import ceil, floor
from unittest import TestCase


from musicscore import Metronome
from musurgia.timing.duration import (
    Duration,
    convert_duration_to_quarter_duration_value,
    convert_quarter_duration_to_duration_value,
)


class TestDuration(TestCase):
    def test_seconds(self):
        d = Duration(seconds=10)
        assert d.seconds == 10
        assert d.minutes == 0
        assert d.hours == 0
        assert d.get_clock_as_string() == "0:00:10.0"

    def test_seconds_over_60(self):
        d = Duration(seconds=70)
        assert d.seconds == 10
        assert d.minutes == 1
        assert d.hours == 0
        assert d.get_clock_as_string() == "0:01:10.0"

    def test_seconds_over_3600(self):
        d = Duration(seconds=3610)
        assert d.seconds == 10
        assert d.minutes == 0
        assert d.hours == 1
        assert d.get_clock_as_string() == "1:00:10.0"

    def test_seconds_float(self):
        d = Duration(seconds=10.5)
        assert d.get_clock_as_string() == "0:00:10.5"

    def test_minutes(self):
        d = Duration(minutes=10)
        assert d.seconds == 0
        assert d.minutes == 10
        assert d.hours == 0
        assert d.get_clock_as_string() == "0:10:00.0"

    def test_minutes_over_60(self):
        d = Duration(minutes=75)
        assert d.seconds == 0
        assert d.minutes == 15
        assert d.hours == 1
        assert d.get_clock_as_string() == "1:15:00.0"

    def test_minutes_float(self):
        d = Duration(minutes=10.5)
        assert d.get_clock_as_string() == "0:10:30.0"

    def test_hours(self):
        d = Duration(hours=3)
        assert d.seconds == 0
        assert d.minutes == 0
        assert d.hours == 3
        assert d.get_clock_as_string() == "3:00:00.0"

    def test_no_arguments(self):
        d = Duration()
        assert d.seconds == 0
        assert d.minutes == 0
        assert d.hours == 0
        assert d.get_clock_as_string() == "0:00:00.0"

    def test_complex_input(self):
        d = Duration(hours=2.5, minutes=70.5, seconds=70.5)
        assert d.get_clock_as_string() == "3:41:40.5"

    def test_get_clock_modes(self):
        d = Duration(hours=2.5, minutes=90.5, seconds=90.5)
        assert d.get_clock_as_string(mode="hms") == "4:02:00.5"
        assert d.get_clock_as_string(mode="ms") == "02:00.5"
        assert d.get_clock_as_string(mode="msreduced") == "2:0.5"

    def test_calculate_in_seconds(self):
        d = Duration(hours=1, minutes=30, seconds=30)
        assert d.calculate_in_seconds() == 5430.0

    def test_calculate_in_minutes(self):
        d = Duration(hours=1, minutes=30, seconds=30)
        assert d.calculate_in_minutes() == 90.5

    def test_calculate_in_hours(self):
        d = Duration(hours=1, minutes=30, seconds=30)
        assert float(d.calculate_in_hours()) == 1.5083333333333333

    def test_set_and_get_values(self):
        d = Duration(hours=1, minutes=30, seconds=30)
        assert d.clock.get_values() == (1, 30, 30)
        d.hours = 2
        assert d.clock.get_values() == (2, 30, 30)
        d.minutes = 10
        assert d.clock.get_values() == (2, 10, 30)
        d.seconds = 20.0
        assert d.clock.get_values() == (2, 10, 20)
        d.minutes = 65
        assert d.clock.get_values() == (3, 5, 20)

    def test_fractioned_clock(self):
        d = Duration(Fraction(10, 3))
        assert d.seconds == Fraction(10, 3)


# class TestConvertors(TestCase):
#     def test_convert_duration_to_quarter_duration(self):
#         t = 60
#         d = Duration(seconds=3)
#         assert convert_duration_to_quarter_duration_value(d, t) == 3
#         t = 30
#         assert convert_duration_to_quarter_duration_value(d, t) == QuarterDuration(1.5)
#         assert convert_duration_to_quarter_duration_value(3, 120) == 6
#         t = Metronome(60, 2)
#         assert convert_duration_to_quarter_duration_value(3, t) == 6

#     import unittest


class QuarterDurationConversionTestCase(TestCase):
    def test_basic_case(self):
        # Test case 1: 60 quarters per minute, 60 seconds duration
        metronome = Metronome(per_minute=60, beat_unit=1)  # 60 quarters per minute
        duration = 60  # 60 seconds
        self.assertEqual(
            convert_duration_to_quarter_duration_value(metronome, duration), 60
        )

    def test_half_quarters(self):
        # Test case 2: 120 quarters per minute, 30 seconds duration
        metronome = Metronome(per_minute=120, beat_unit=1)  # 120 quarters per minute
        duration = 30  # 30 seconds
        self.assertEqual(
            convert_duration_to_quarter_duration_value(metronome, duration), 60
        )

    def test_triplet_quarters(self):
        # Test case 3: 90 quarters per minute, 20 seconds duration
        metronome = Metronome(per_minute=90, beat_unit=1)  # 90 quarters per minute
        duration = 20  # 20 seconds
        self.assertEqual(
            convert_duration_to_quarter_duration_value(metronome, duration), 30
        )

    def test_fractional_duration(self):
        # Test case 4: 120 quarters per minute, 2.5 seconds duration
        metronome = Metronome(per_minute=120, beat_unit=1)  # 120 quarters per minute
        duration = 2.5  # 2.5 seconds
        self.assertEqual(
            convert_duration_to_quarter_duration_value(metronome, duration), 5
        )

    def test_zero_duration(self):
        # Test case 5: 120 quarters per minute, 0 seconds duration
        metronome = Metronome(per_minute=120, beat_unit=1)  # 120 quarters per minute
        duration = 0  # 0 seconds
        self.assertEqual(
            convert_duration_to_quarter_duration_value(metronome, duration), 0
        )

    def test_high_bpm_short_duration(self):
        # Test case 6: 240 quarters per minute, 0.5 seconds duration
        metronome = Metronome(per_minute=240, beat_unit=1)  # 240 quarters per minute
        duration = 0.5  # 0.5 seconds
        self.assertEqual(
            convert_duration_to_quarter_duration_value(metronome, duration), 2
        )

    def test_low_bpm_long_duration(self):
        # Test case 7: 30 quarters per minute, 120 seconds duration
        metronome = Metronome(per_minute=30, beat_unit=1)  # 30 quarters per minute
        duration = 120  # 120 seconds
        self.assertEqual(
            convert_duration_to_quarter_duration_value(metronome, duration), 60
        )

    def test_edge_case_small_bpm(self):
        # Test case 8: 15 quarters per minute, 60 seconds duration
        metronome = Metronome(per_minute=15, beat_unit=1)  # 15 quarters per minute
        duration = 60  # 60 seconds
        self.assertEqual(
            convert_duration_to_quarter_duration_value(metronome, duration), 15
        )

    def test_edge_case_large_bpm(self):
        # Test case 9: 480 quarters per minute, 1 second duration
        metronome = Metronome(per_minute=480, beat_unit=1)  # 480 quarters per minute
        duration = 1  # 1 second
        self.assertEqual(
            convert_duration_to_quarter_duration_value(metronome, duration), 8
        )

    def test_exact_quarter_duration(self):
        # Test case 10: 60 quarters per minute, 1 second duration
        metronome = Metronome(per_minute=60, beat_unit=1)  # 60 quarters per minute
        duration = 1  # 1 second
        self.assertEqual(
            convert_duration_to_quarter_duration_value(metronome, duration), 1
        )


class DurationToQuarterDurationConversionTestCase(TestCase):
    def test_basic_case(self):
        # Test case 1: Convert 60 quarters to seconds with metronome 120 BPM and beat unit 1 (quarter note per beat)
        metronome = Metronome(per_minute=120, beat_unit=1)
        quarter_duration = 60  # 60 quarters
        self.assertEqual(
            convert_quarter_duration_to_duration_value(metronome, quarter_duration), 30
        )

    def test_half_note_case(self):
        # Test case 2: Convert 60 quarters to seconds with metronome 120 BPM and beat unit 2 (half note per beat)
        metronome = Metronome(
            per_minute=120, beat_unit=2
        )  # 2 quarters per beat (half note)
        quarter_duration = 60  # 60 quarters
        self.assertEqual(
            convert_quarter_duration_to_duration_value(metronome, quarter_duration), 15
        )

    def test_eighth_note_case(self):
        # Test case 3: Convert 15 quarters to seconds with metronome 90 BPM and beat unit 0.5 (eighth note per beat)
        metronome = Metronome(
            per_minute=90, beat_unit=0.5
        )  # 0.5 quarters per beat (eighth note)
        quarter_duration = 15  # 15 quarters
        self.assertEqual(
            convert_quarter_duration_to_duration_value(metronome, quarter_duration), 20
        )

    def test_zero_quarter_duration(self):
        # Test case 4: Convert 0 quarters to seconds (edge case)
        metronome = Metronome(per_minute=120, beat_unit=1)  # 1 quarter per beat
        quarter_duration = 0  # 0 quarters
        self.assertEqual(
            convert_quarter_duration_to_duration_value(metronome, quarter_duration), 0
        )

    def test_fractional_quarters(self):
        # Test case 5: Convert 5 quarters to seconds with metronome 60 BPM and beat unit 0.5 (eighth note per beat)
        metronome = Metronome(
            per_minute=60, beat_unit=0.5
        )  # 0.5 quarters per beat (eighth note)
        quarter_duration = 5  # 5 quarters
        self.assertEqual(
            convert_quarter_duration_to_duration_value(metronome, quarter_duration), 10
        )

    def test_large_bpm_case(self):
        # Test case 6: Convert 120 quarters to seconds with metronome 240 BPM and beat unit 1 (quarter note per beat)
        metronome = Metronome(per_minute=240, beat_unit=1)  # 1 quarter per beat
        quarter_duration = 120  # 120 quarters
        self.assertEqual(
            convert_quarter_duration_to_duration_value(metronome, quarter_duration), 30
        )

    def test_small_bpm_case(self):
        # Test case 7: Convert 10 quarters to seconds with metronome 30 BPM and beat unit 1 (quarter note per beat)
        metronome = Metronome(per_minute=30, beat_unit=1)  # 1 quarter per beat
        quarter_duration = 10  # 10 quarters
        self.assertEqual(
            convert_quarter_duration_to_duration_value(metronome, quarter_duration), 20
        )

    def test_large_quarter_duration(self):
        # Test case 8: Convert 500 quarters to seconds with metronome 60 BPM and beat unit 1 (quarter note per beat)
        metronome = Metronome(per_minute=60, beat_unit=1)  # 1 quarter per beat
        quarter_duration = 500  # 500 quarters
        self.assertEqual(
            convert_quarter_duration_to_duration_value(metronome, quarter_duration), 500
        )

    def test_fractional_bpm_case(self):
        # Test case 9: Convert 40 quarters to seconds with metronome 75 BPM and beat unit 1 (quarter note per beat)
        metronome = Metronome(per_minute=75, beat_unit=1)  # 1 quarter per beat
        quarter_duration = 40  # 40 quarters
        self.assertEqual(
            convert_quarter_duration_to_duration_value(metronome, quarter_duration), 32
        )

    def test_edge_case_high_beat_unit(self):
        # Test case 10: Convert 10 quarters to seconds with metronome 120 BPM and beat unit 2 (half note per beat)
        metronome = Metronome(
            per_minute=120, beat_unit=2
        )  # 2 quarters per beat (half note)
        quarter_duration = 10  # 10 quarters
        self.assertEqual(
            convert_quarter_duration_to_duration_value(metronome, quarter_duration), 2.5
        )


class TestMagics(TestCase):
    cl = Duration

    def setUp(self):
        self.main = self.cl(70)
        self.equal = self.cl(70)
        self.equal_float = 70.0
        self.larger = self.cl(80)
        self.larger_float = 80.0
        self.smaller = self.cl(60)
        self.smaller_float = 60.0

    def test_abs(self):
        assert abs(self.cl(-70)).calculate_in_seconds() == 70

    def test_ceil(self):
        assert ceil(self.cl(70.2)).calculate_in_seconds() == 71

    def test_floor(self):
        assert floor(self.cl(70.2)).calculate_in_seconds() == 70

    def test_floor_division(self):
        a = self.cl(10)
        b = self.cl(4)
        c = self.cl(2)
        assert a // b == c
        assert a // 4 == c
        assert a // b == 2
        assert a // 4 == 2

    def test_gt(self):
        assert self.main > self.smaller
        assert self.main > self.smaller_float
        assert not self.main > self.equal
        assert not self.main > self.equal_float
        assert not self.main > self.larger
        assert not self.main > self.larger_float

    def test_ge(self):
        assert self.main >= self.smaller
        assert self.main >= self.smaller_float
        assert self.main >= self.equal
        assert self.main >= self.equal_float
        assert not self.main >= self.larger
        assert not self.main >= self.larger_float

    def test_le(self):
        assert not self.main <= self.smaller
        assert not self.main <= self.smaller_float
        assert self.main <= self.equal
        assert self.main <= self.equal_float
        assert self.main <= self.larger
        assert self.main <= self.larger_float

    def test_lt(self):
        assert not self.main < self.smaller
        assert not self.main < self.smaller_float
        assert not self.main < self.equal
        assert not self.main < self.equal_float
        assert self.main < self.larger
        assert self.main < self.larger_float

    def test_mod(self):
        a = self.cl(10)
        b = self.cl(3)
        c = self.cl(1)
        assert a % 3 == c
        assert a % 3 == 1
        assert a % b == c
        assert a % b == 1

    def test_mul(self):
        a = self.cl(10)
        b = self.cl(3)
        c = self.cl(30)
        assert a * b == 30
        assert a * b == c
        assert a * 3 == 30
        assert a * 3 == c

    def test_neg(self):
        a = self.cl(10)
        b = self.cl(-10)
        assert -a == b
        assert -a == -10
        assert -b == a
        assert -b == 10

    def test_pos(self):
        a = self.cl(10)
        assert +a == a

    def test_power(self):
        a = self.cl(10)
        b = self.cl(100)
        assert 10.0**2 == 100
        assert a**2 == 100
        assert a**2 == b

    def test_radd(self):
        a = self.cl(10)
        b = self.cl(100)
        assert a.__radd__(b) == b + a

    def test_rmod(self):
        a = self.cl(3)
        b = self.cl(10)
        assert a.__rmod__(b) == b % a

    def test_rmul(self):
        a = self.cl(3)
        b = self.cl(10)
        assert a.__rmul__(b) == b * a

    def test_eq(self):
        a = self.cl(10)
        b = self.cl(10)
        c = self.cl(11)
        assert a == b
        assert a == 10
        assert 10 == a
        assert a == 10.0
        assert a != 11
        assert a != c
        assert not a == c
        assert not a == 11
        assert not 11 == a
        d = self.cl(Fraction(10, 3))
        assert Fraction(10, 3) == Fraction(10, 3)
        assert Fraction(10, 3).__eq__(Fraction(10, 3))
        assert d.__eq__(Fraction(10, 3))
        assert d == Fraction(10, 3)
        assert a is not None

    def test_round(self):
        assert self.cl(70.7) == self.cl(70.7)
        assert round(self.cl(70.67), 1) == 70.7
        assert round(self.cl(70.67), 1) == self.cl(70.7)
        assert round(self.cl(70.67), 1) != self.cl(70.6)
        assert round(self.cl(70.67), 1) != 70.6

    def test_rtruediv(self):
        a = self.cl(3)
        b = self.cl(10)
        assert a.__rtruediv__(b) == Fraction(10, 3)

    def test_truediv(self):
        a = self.cl(10)
        b = self.cl(3)
        assert a / b == Fraction(10, 3)

    def test_trunc(self):
        a = self.cl(10.233)
        assert a.__trunc__() == 10

    def test_rfloordiv(self):
        a = self.cl(3)
        b = self.cl(10)
        assert a.__rfloordiv__(b) == b // a



from typing import Any

from musurgia.musurgia_types import check_type, PdfUnitType

__all__ = []  # type: ignore


class PdfUnitTypeCheckMeta(type):
    def __setattr__(cls, key: str, value: Any) -> None:
        if key == "GLOBAL_UNIT":
            check_type(
                value,
                "PdfUnitType",
                class_name="PdfUnit",
                class_attribute_name="GLOBAL_UNIT",
            )
        super().__setattr__(key, value)


class PdfUnit(metaclass=PdfUnitTypeCheckMeta):
    _DEFAULT_UNIT = "mm"
    GLOBAL_UNIT = _DEFAULT_UNIT

    @staticmethod
    def get_k() -> float:
        k_dict = {"pt": 1.0, "mm": 72 / 25.4, "cm": 72 / 2.54, "in": 72.0}
        k = k_dict[PdfUnit.GLOBAL_UNIT]
        return k

    @staticmethod
    def reset() -> None:
        PdfUnit.GLOBAL_UNIT = PdfUnit._DEFAULT_UNIT

    @staticmethod
    def change(val: PdfUnitType) -> None:
        PdfUnit.GLOBAL_UNIT = val



from abc import ABC
from typing import Any, cast, Optional, Callable

from musurgia.musurgia_exceptions import (
    RulerCannotSetLengthsError,
    RulerLengthNotPositiveError,
    TimeRulerCannotSetLength,
)
from musurgia.musurgia_types import ConvertibleToFloat, check_type, ClockMode
from musurgia.pdf import TextLabel
from musurgia.pdf.line import (
    AbstractSegmentedLine,
    MarkLine,
    VerticalSegmentedLine,
    HorizontalSegmentedLine,
)
from musurgia.timing.duration import Duration

__all__ = ["HorizontalRuler", "VerticalRuler", "TimeRuler"]


class AbstractRuler(AbstractSegmentedLine, ABC):
    def __init__(
        self,
        length: ConvertibleToFloat,
        unit: ConvertibleToFloat = 10.0,
        first_label: int = 0,
        label_show_interval: int = 1,
        show_first_label: bool = True,
        *args: Any,
        **kwargs: Any,
    ):
        if "lengths" in kwargs:
            raise RulerCannotSetLengthsError
        check_type(
            length,
            "ConvertibleToFloat",
            class_name="AbstractRuler",
            property_name="length",
        )
        length = float(length)
        if length < 0:
            raise RulerLengthNotPositiveError
        check_type(
            unit, "ConvertibleToFloat", class_name="AbstractRuler", property_name="unit"
        )
        check_type(
            first_label, int, class_name="AbstractRuler", property_name="first_label"
        )
        check_type(
            label_show_interval,
            int,
            class_name="AbstractRuler",
            property_name="label_show_interval",
        )
        check_type(
            show_first_label,
            bool,
            class_name="AbstractRuler",
            property_name="show_first_label",
        )

        unit = float(unit)
        number_of_units = float(length) / unit
        partial_segment_length = number_of_units - int(number_of_units)
        lengths = int(number_of_units) * [unit]
        if partial_segment_length:
            lengths += [partial_segment_length * unit]
        super().__init__(lengths=lengths, *args, **kwargs)  # type: ignore
        if partial_segment_length:
            self.segments[-1].end_mark_line.show = False

        self._unit = unit
        self._first_label = first_label
        self._label_show_interval = label_show_interval
        self._show_first_label = show_first_label
        self._set_labels()

    def _set_labels(self) -> None:
        def _add_label(mark_line: MarkLine, txt: str) -> None:
            tl = TextLabel(txt, master=mark_line)
            if isinstance(self, VerticalSegmentedLine):
                tl.placement = "left"
                tl.right_margin = 1
                tl.top_margin = 0
            else:
                tl.bottom_margin = 1
            mark_line.add_text_label(tl)

        for index, segment in enumerate(self.segments):
            if not self.get_show_first_label() and index == 0:
                pass
            else:
                if index % self.get_label_show_interval() == 0:
                    mark_line = segment.start_mark_line
                    _add_label(mark_line, str(index + self.get_first_label()))

        if self.segments:
            last_segment_end_mark_line = self.segments[-1].end_mark_line
            if (
                last_segment_end_mark_line.show
                and (len(self.segments)) % self.get_label_show_interval() == 0
            ):
                _add_label(
                    last_segment_end_mark_line,
                    str(len(self.segments) + self.get_first_label()),
                )

    def get_markline_text_labels(self) -> list[TextLabel]:
        return [
            label
            for seg in self.segments
            for label in seg.start_mark_line.get_text_labels()
            + seg.end_mark_line.get_text_labels()
        ]

    def change_labels(
        self, condition: Optional[Callable[[TextLabel], bool]] = None, **kwargs: Any
    ) -> None:
        if condition is None:
            condition = lambda label: True  # noqa
        labels = [l for l in self.get_markline_text_labels() if condition(l)]
        for label in labels:
            for key, value in kwargs.items():
                setattr(label, key, value)

    @property
    def length(self) -> None:
        raise AttributeError("use get_length() instead")

    @length.setter
    def length(self, val: Any) -> None:
        raise AttributeError("length is not settable after initialization.")

    @property
    def unit(self) -> None:
        raise AttributeError("use get_unit() instead")

    @unit.setter
    def unit(self, val: Any) -> None:
        raise AttributeError("unit is not settable after initialization.")

    @property
    def first_label(self) -> None:
        raise AttributeError("use get_first_label() instead")

    @first_label.setter
    def first_label(self, val: Any) -> None:
        raise AttributeError("first_label is not settable after initialization.")

    @property
    def label_show_interval(self) -> None:
        raise AttributeError("use get_label_show_interval() instead")

    @label_show_interval.setter
    def label_show_interval(self, val: Any) -> None:
        raise AttributeError(
            "label_show_interval is not settable after initialization."
        )

    @property
    def show_first_label(self) -> None:
        raise AttributeError("use get_show_first_label() instead")

    @show_first_label.setter
    def show_first_label(self, val: Any) -> None:
        raise AttributeError("show_first_label is not settable after initialization.")

    def get_unit(self) -> float:
        return float(self._unit)

    def get_length(self) -> float:
        return float(sum([seg.length for seg in self.segments]))

    def get_first_label(self) -> int:
        return self._first_label

    def get_label_show_interval(self) -> int:
        return self._label_show_interval

    def get_show_first_label(self) -> bool:
        return self._show_first_label


class HorizontalRuler(AbstractRuler, HorizontalSegmentedLine):
    pass


class VerticalRuler(AbstractRuler, VerticalSegmentedLine):
    pass


class TimeRuler(HorizontalRuler):
    def __init__(
        self,
        duration: int,
        unit: ConvertibleToFloat = 2,
        label_show_interval: int = 10,
        shrink_factor: ConvertibleToFloat = 0.6,
        mark_line_size: ConvertibleToFloat = 4,
        clock_mode: ClockMode = "hms",
        *args: Any,
        **kwargs: Any,
    ):
        if "length" in kwargs:
            raise TimeRulerCannotSetLength
        check_type(
            duration,
            "PositiveInteger",
            class_name="TimeRuler",
            property_name="duration",
        )
        super().__init__(
            length=duration * unit,
            unit=unit,
            label_show_interval=label_show_interval,
            *args,
            **kwargs,
        )  # type: ignore
        self._clock_mode: ClockMode = clock_mode
        self._change_label_texts()
        self._shrink_factor: ConvertibleToFloat
        self._mark_line_size: ConvertibleToFloat
        self.shrink_factor = shrink_factor  # type: ignore
        self.mark_line_size = mark_line_size  # type: ignore

    def _change_label_texts(self) -> None:
        for label in self.get_markline_text_labels():
            duration = Duration(float(label.value))
            label.value = duration.get_clock_as_string(mode=self._clock_mode, round_=1)
            if label.value[-2:] == ".0":
                label.value = label.value[:-2]

    def _change_mark_line_lengths(self) -> None:
        try:
            sf = self.shrink_factor
            mls = self.mark_line_size
            for i, seg in enumerate(self.segments):
                if i % self.get_label_show_interval() == 0:
                    seg.start_mark_line.length = mls
                else:
                    seg.start_mark_line.length = mls * sf

            if len(self.segments) % self.get_label_show_interval() == 0:
                self.segments[-1].end_mark_line.length = mls
            else:
                self.segments[-1].end_mark_line.length = mls * sf

        except AttributeError:
            pass

    def get_duration(self) -> float:
        return self.get_length() / self.get_unit()

    @property
    def shrink_factor(self) -> float:
        return cast(float, self._shrink_factor)

    @shrink_factor.setter
    def shrink_factor(self, val: ConvertibleToFloat) -> None:
        self._shrink_factor = float(val)
        self._change_mark_line_lengths()

    @property
    def mark_line_size(self) -> float:
        return cast(float, self._mark_line_size)

    @mark_line_size.setter
    def mark_line_size(self, val: ConvertibleToFloat) -> None:
        self._mark_line_size = float(val)
        self._change_mark_line_lengths()



from pathlib import Path
from typing import Any

import matplotlib as mpl
from matplotlib._afm import AFM
from musurgia.musurgia_types import FontFamily, FontWeight, FontStyle, check_type

__all__ = ["Font"]


def _make_afm_path_dictionary() -> "dict[tuple[Any, Any, str], AFM]":
    def check_entry() -> bool:
        old_afm = output.get((family, weight, style))
        if old_afm is not None:
            old_header = old_afm._header
            new_header = afm._header
            diff = set(old_header) ^ set(new_header)
            if diff == set():
                return False
            else:
                return True
            # if diff == {b'CapHeight'}:
            #     if new_header.get(b'CapHeight'):
            #         return True
            # elif diff == set():
            #     return False
            # else:
            #     raise AttributeError(
            #         f'{family}, {weight}, {style} already in dict: {old_afm} difference: {diff}')
        else:
            return True

    output = {}
    directory = Path(mpl.get_data_path(), "fonts", "afm")
    for file in directory.iterdir():
        afm_path = file
        with afm_path.open("rb") as fh:
            afm = AFM(fh)  # type: ignore
        family = afm.get_familyname()  # type: ignore
        weight = afm.get_weight().lower()  # type: ignore
        if afm.get_angle() < 0:  # type: ignore
            style = "italic"
        else:
            style = "regular"
        if check_entry():
            output[family, weight, style] = afm

    return output


class Font:
    """
    Class representing a font. It is used in class Text to set font family, weight, style and size.
    It accesses matplotlib._afm.AFM class (Adobe Font Metrics) to be able to determine the exact height and width of
    the text in pixels via two methods: :obj:`get_text_pixel_height()` and :obj:`get_text_pixel_width()`

    Attributes:
        family: Default value is ``Courier``.
        weight: Default value is ``medium``
        style: Default value is ``regular``
        size: Default value is `10`
    """

    __AFM_PATH_DICTIONARY = _make_afm_path_dictionary()

    # pprint(__AFM_PATH_DICTIONARY)
    # _FAMILY = ['Courier']
    # _WEIGHT = ['bold', 'medium']
    # _STYLE = ['italic', 'regular']

    def __init__(
        self,
        family: FontFamily = "Courier",
        weight: FontWeight = "medium",
        style: FontStyle = "regular",
        size: int = 10,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._family: FontFamily
        self._weight: FontWeight
        self._style: FontStyle
        self._size: int
        self._afm: AFM

        self.family = family
        self.weight = weight
        self.style = style
        self.size = size

    def _set_afm(self) -> None:
        self._afm = self.__AFM_PATH_DICTIONARY[self.family, self.weight, self.style]

    @property
    def family(self) -> FontFamily:
        """
        Set and get font family. Currently valid values are [``Helvetica``, ``Courier``, ``Times``]
        """
        return self._family

    @family.setter
    def family(self, val: FontFamily) -> None:
        check_type(
            val,
            "FontFamily",
            class_name=self.__class__.__name__,
            property_name="family",
        )
        self._family = val
        try:
            self._set_afm()
        except AttributeError:
            pass

    @property
    def size(self) -> int:
        """
        Set and get font size
        """
        return self._size

    @size.setter
    def size(self, val: int) -> None:
        check_type(val, int, class_name=self.__class__.__name__, property_name="size")
        self._size = val

    @property
    def style(self) -> FontStyle:
        """
        Set and get font style. Valid values are [``italic``, ``regular``]
        """
        return self._style

    @style.setter
    def style(self, val: FontStyle) -> None:
        check_type(
            val, "FontStyle", class_name=self.__class__.__name__, property_name="style"
        )
        self._style = val
        try:
            self._set_afm()
        except AttributeError:  # pragma: no cover
            pass

    @property
    def weight(self) -> FontWeight:
        """
        Set and get font weight. Valid values are [``bold``, ``medium``]
        """
        return self._weight

    @weight.setter
    def weight(self, val: FontWeight) -> None:
        check_type(
            val,
            "FontWeight",
            class_name=self.__class__.__name__,
            property_name="weight",
        )
        self._weight = val
        try:
            self._set_afm()
        except AttributeError:
            pass

    def get_text_pixel_width(self, value: str) -> float:
        """
        :param value: text as str
        :return: width of text in pixels

        >>> round(Font().get_text_pixel_width('Test'), 2)
        24.0
        >>> round(Font(size=12).get_text_pixel_width('Test'), 2)
        28.8
        """
        return (self._afm.string_width_height(value)[0] / 1000) * self.size  # type: ignore

    def get_text_pixel_height(self, val: str) -> float:
        """
        :param val: text as str
        :return: height of text in pixels

        >>> round(Font().get_text_pixel_height('Test'), 2)
        5.77
        >>> round(Font(size=12).get_text_pixel_height('Test'), 2)
        6.92
        >>> round(Font(size=12, weight='bold', style='italic').get_text_pixel_height('Test'), 2)
        6.95
        """
        return (self._afm.string_width_height(val)[1] / 1000) * self.size  # type: ignore



from pathlib import Path

from musurgia.pdf import Pdf, StraightLine, draw_ruler
from musurgia.tests.utils_for_tests import (
    PdfTestCase,
    add_control_positions_to_draw_object,
    add_test_labels,
)

path = Path(__file__)


class TestPdfUtil(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.line = StraightLine(length=100, mode="h", relative_x=10, relative_y=10)
        self.line.margins = (5, 10, 15, 20)

    def test_add_control_positions(self):
        add_control_positions_to_draw_object(self.line)

        with self.file_path(path, "add_control_positions", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "v")
            draw_ruler(self.pdf, "h")
            self.pdf.translate(10, 10)
            self.line.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_add_test_labels(self):
        add_test_labels(self.line)
        with self.file_path(path, "add_test_labels", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            # draw_ruler(self.pdf, 'v')
            # draw_ruler(self.pdf, 'h')
            self.pdf.translate(10, 10)
            self.line.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)



from abc import abstractmethod
from fractions import Fraction
import warnings
from verysimpletree.tree import Tree
from typing import Any, TypeVar, Union
from musurgia.musurgia_exceptions import WrongTreeValueError, WrongTreeValueWarning
from musurgia.musurgia_types import ConvertibleToFraction

T = TypeVar("T", bound="ValuedTree")


class ValuedTree(Tree[Any]):
    def _change_children_value(self, factor: Union[int, float, Fraction]) -> None:
        for child in self._get_children():
            child._set_value(child.get_value() * factor)
            child._change_children_value(factor)

    def _check_tree_children(self) -> None:
        _children = super().get_children()
        self._check_tree_children_values([ch.get_value() for ch in _children])

    def _check_tree_children_values(
        self, children_values: list[ConvertibleToFraction]
    ) -> None:
        if sum(children_values) != self.get_value():
            raise WrongTreeValueError(
                f"Children of ValuedTree node of position {self.get_position_in_tree()} with value {self.get_value()} have wrong values {children_values} (sum={sum(children_values)})"
            )

    def _get_children(self: T) -> list[T]:
        return super().get_children()

    @abstractmethod
    def _set_value(self, val: ConvertibleToFraction) -> None:
        """_set_value must be defined."""

    @property
    def value(self) -> None:
        raise AttributeError("Use get_value() instead.")

    def check_tree_values(self) -> bool:
        for node in self.traverse():
            if not node.is_leaf:
                node._check_tree_children()
        return True

    @abstractmethod
    def get_value(self) -> Fraction:
        """get_value must be defined."""

    def update_value(self, new_value: ConvertibleToFraction) -> None:
        if not isinstance(new_value, Fraction):
            new_value = Fraction(new_value)
        factor = Fraction(new_value, self.get_value())
        self._set_value(new_value)
        for node in self.get_reversed_path_to_root()[1:]:
            node._set_value(sum([child.get_value() for child in node._get_children()]))

        self._change_children_value(factor)

    def get_children(self: T) -> list[T]:
        children = self._get_children()
        if not self.is_leaf:
            try:
                self._check_tree_children_values([ch.get_value() for ch in children])
            except WrongTreeValueError as err:
                warnings.warn(str(err), WrongTreeValueWarning, stacklevel=2)
        return children



from pathlib import Path

from musurgia.pdf.line import (
    HorizontalLineSegment,
    HorizontalSegmentedLine,
    VerticalSegmentedLine,
)
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.pdf_tools import draw_ruler
from musurgia.tests.utils_for_tests import PdfTestCase

path = Path(__file__)


class TestMarkLineLabels(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.ls = HorizontalLineSegment(length=20)
        self.ls.set_straight_line_relative_y(0)

    def test_draw_above(self):
        ml = self.ls.start_mark_line
        ml.add_text_label("first text label above")
        ml.add_text_label("second text label above")
        ml.add_text_label("third  text label above")
        self.ls.relative_x = 10
        self.ls.relative_y = 20
        with self.file_path(path, "above", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")

            self.ls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_draw_one_above(self):
        ml = self.ls.start_mark_line
        ml.length = 20
        ml.add_text_label("first text label above")

        with self.file_path(path, "one_above", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.translate(30, 30)
            self.ls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_draw_below(self):
        ml = self.ls.start_mark_line
        ml.add_text_label("first text label below", placement="below")
        ml.add_text_label("second text label below", placement="below")
        self.ls.relative_x = 10
        self.ls.relative_y = 20
        with self.file_path(path, "below", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.ls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_draw_left(self):
        ml = self.ls.start_mark_line
        ml.add_text_label("first text label left", placement="left")
        ml.add_text_label("second text label left left left", placement="left")
        ml.add_text_label("third text label left left left", placement="left")
        ml.get_left_text_labels()[1].font.size = 8
        self.ls.relative_x = 40
        self.ls.relative_y = 10
        with self.file_path(path, "left", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.ls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_different_sizes(self):
        ml = self.ls.start_mark_line
        ml.add_text_label("first text label above", font_size=7)
        ml.add_text_label("second text label above", font_size=8)
        ml.add_text_label("third text label above", font_size=9)
        with self.file_path(path, "different_sizes", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(10, 20)
            self.ls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_font_size_8(self):
        ml = self.ls.start_mark_line

        ml.add_text_label("first text label above", font_size=8, bottom_margin=2)
        ml.add_text_label("second text label above", font_size=8, bottom_margin=4)
        ml.add_text_label("third text label above", font_size=8)
        with self.file_path(path, "font_size_8", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(10, 20)
            self.ls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_below_with_different_mark_line_lengths(self):
        hsl = HorizontalSegmentedLine(lengths=[10, 15, 20])
        hsl.segments[0].start_mark_line.length = 6
        hsl.segments[0].start_mark_line.add_label(1, placement="below", font_size=8)
        hsl.segments[1].start_mark_line.add_label(2, placement="below", font_size=8)
        hsl.segments[2].start_mark_line.add_label(3, placement="below", font_size=8)
        hsl.set_straight_line_relative_y(0)
        with self.file_path(
            path, "below_with_different_mark_line_lengths", "pdf"
        ) as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(10, 10)
            hsl.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_vertical_with_left_text_labels(self):
        vsl = VerticalSegmentedLine(lengths=[10, 15, 20])
        vsl.segments[0].start_mark_line.length = 6
        vsl.segments[0].start_mark_line.add_label(1, placement="left", font_size=8)
        vsl.segments[1].start_mark_line.add_label(2, placement="left", font_size=8)
        vsl.segments[2].start_mark_line.add_label(3, placement="left", font_size=8)
        with self.file_path(path, "vertical_with_left_text_labels", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(10, 10)
            vsl.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_draw_above_with_different_bottom_margins(self):
        ml = self.ls.start_mark_line
        ml.add_text_label("first text label above", bottom_margin=2)
        ml.add_text_label("second text label above", bottom_margin=4)
        ml.add_text_label("third  text label above", bottom_margin=15)
        self.ls.relative_x = 10
        self.ls.relative_y = 40
        with self.file_path(
            path, "above_with_different_bottom_margins", "pdf"
        ) as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")

            self.ls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_left_position(self):
        ml = self.ls.start_mark_line
        ml.add_text_label("left one", placement="left")
        ml.add_text_label("left two", placement="left")
        ml.add_text_label("left three", placement="left")
        ml.add_text_label("left four", placement="left")

        with self.file_path(path, "left_position", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(30, 30)
            self.ls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)



from pathlib import Path
from unittest import TestCase

from musicscore.score import Score

from musurgia.tests.utils_for_tests import XMLTestCase, create_test_valued_tree
from musurgia.trees.timelinetree import (
    SimpleTimelineChordFactory,
    TimelineDuration,
    TimelineTree,
)
from musicscore.chord import Chord
from musicscore.metronome import Metronome


class TimelineChordTreeTestCase(TestCase):
    def setUp(self):
        self.tlt = TimelineTree(duration=TimelineDuration(4))
        self.chf = SimpleTimelineChordFactory(self.tlt, show_metronome=True)

    def test_tree_chord(self):
        ch = self.chf.create_chord()
        self.assertTrue(isinstance(ch, Chord))
        with self.assertRaises(AttributeError):
            self.chf.chord = Chord(60, 2)
        self.assertEqual(ch.metronome.per_minute, 60)
        self.assertEqual(ch.quarter_duration, self.tlt.get_value())
        self.assertEqual(ch.midis[0].value, 72)

    def test_update_quarter_duration(self):
        self.assertEqual(self.tlt.get_duration().get_quarter_duration(), 4)
        self.tlt.get_duration().metronome = Metronome(120)
        self.assertEqual(self.chf.create_chord().quarter_duration, 8)
        self.tlt.update_duration(5)
        self.assertEqual(self.chf.create_chord().quarter_duration, 10)
        self.tlt.get_duration().metronome = Metronome(120, 2)
        self.tlt.metronome = Metronome(120, 2)
        self.assertEqual(self.chf.create_chord().quarter_duration, 20)


path = Path(__file__)


class TimeLineTreeToScoreTestCase(XMLTestCase):
    def setUp(self):
        self.vt = create_test_valued_tree()
        self.tmt = TimelineTree.create_tree_from_list(
            tree_list_representation=self.vt.get_list_representation(
                key=lambda node: node.get_value()
            ),
            represented_attribute_names="duration",
        )
        self.score = Score()

    def test_timeline_tree_layers_to_score(self):
        _show_metronome = True
        for layer_number in range(self.tmt.get_number_of_layers() + 1):
            part = self.score.add_part(f"part-{layer_number + 1}")
            layer = self.tmt.get_layer(level=layer_number)
            for node in layer:
                part.add_chord(
                    SimpleTimelineChordFactory(
                        node, show_metronome=_show_metronome
                    ).create_chord()
                )
                _show_metronome = False
        self.score.set_possible_subdivisions([2, 3, 4, 5, 6, 7, 8])
        self.score.get_quantized = True
        self.score.finalize()
        with self.file_path(path, "valued_layers_to_score") as xml_path:
            self.score.export_xml(xml_path)



from abc import abstractmethod, ABC
from typing import Any, cast, Union

from musurgia.musurgia_exceptions import (
    SegmentedLineSegmentHasMarginsError,
    SegmentedLineLengthsCannotBeSetError,
)
from musurgia.musurgia_types import (
    HorizontalVertical,
    check_type,
    ConvertibleToFloat,
    MarkLinePlacement,
    PositionType,
    MarginType,
)
from musurgia.pdf.drawobject import (
    SlaveDrawObject,
    MasterDrawObject,
    DrawObject,
    HasShowProtocol,
)
from musurgia.pdf.labeled import Labeled
from musurgia.pdf.margined import Margined
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.positioned import Positioned, HasPositionsProtocol
from musurgia.pdf.rowcolumn import DrawObjectRow, DrawObjectColumn, DrawObjectContainer

__all__ = [
    "HorizontalLineSegment",
    "VerticalLineSegment",
    "StraightLine",
    "HorizontalSegmentedLine",
    "VerticalSegmentedLine",
]


class AbstractStraightLine(Labeled, ABC, HasPositionsProtocol, HasShowProtocol):
    def __init__(
        self,
        mode: HorizontalVertical,
        length: ConvertibleToFloat,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._mode: HorizontalVertical
        self._length: float

        self.mode = mode
        self.length = length  # type: ignore

    @property
    def mode(self) -> HorizontalVertical:
        return self._mode

    @mode.setter
    def mode(self, val: HorizontalVertical) -> None:
        check_type(
            val,
            "HorizontalVertical",
            class_name=self.__class__.__name__,
            property_name="mode",
        )
        self._mode = val

    @property
    def length(self) -> float:
        return self._length

    @length.setter
    def length(self, val: ConvertibleToFloat) -> None:
        check_type(
            val,
            "ConvertibleToFloat",
            class_name=self.__class__.__name__,
            property_name="length",
        )
        self._length = float(val)

    @property
    def is_vertical(self) -> bool:
        if self.mode in ["v", "vertical"]:
            return True
        else:
            return False

    @property
    def is_horizontal(self) -> bool:
        if self.mode in ["h", "horizontal"]:
            return True
        else:
            return False

    @staticmethod
    def get_opposite_mode(mode: HorizontalVertical) -> HorizontalVertical:
        if mode == "h":
            return "v"
        elif mode == "v":
            return "h"
        elif mode == "horizontal":
            return "vertical"
        elif mode == "vertical":
            return "horizontal"
        else:
            raise NotImplementedError  # pragma: no cover

    def get_relative_x2(self) -> float:
        if self.is_horizontal:
            return self.relative_x + self.length
        else:
            return self.relative_x

    def get_relative_y2(self) -> float:
        if self.is_vertical:
            return self.relative_y + self.length
        else:
            return self.relative_y

    def draw(self, pdf: Pdf) -> None:
        if self.show:
            with pdf.pdf_draw_object_translate(cast(DrawObject, self)):
                self.draw_above_text_labels(pdf)
                self.draw_left_text_labels(pdf)
                self.draw_below_text_labels(pdf)
                x2 = self.get_relative_x2() - self.relative_x
                y2 = self.get_relative_y2() - self.relative_y
                pdf.line(0, 0, x2, y2)


class StraightLine(AbstractStraightLine, DrawObject, Positioned, Margined):
    pass


class SlaveStraightLine(AbstractStraightLine, SlaveDrawObject):
    pass


class MarkLine(SlaveStraightLine):
    def __init__(
        self,
        placement: MarkLinePlacement,
        mode: HorizontalVertical,
        length: ConvertibleToFloat = 3,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(length=length, mode=mode, *args, **kwargs)  # type: ignore
        self._placement: MarkLinePlacement
        self.placement = placement

    @property
    def placement(self) -> MarkLinePlacement:
        return self._placement

    @placement.setter
    def placement(self, val: MarkLinePlacement) -> None:
        check_type(
            val,
            "MarkLinePlacement",
            class_name=self.__class__.__name__,
            property_name="placement",
        )
        self._placement = val


class LineSegment(MasterDrawObject, ABC):
    def __init__(
        self,
        mode: HorizontalVertical,
        length: ConvertibleToFloat,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._straight_line = SlaveStraightLine(
            simple_name="straight_line", mode=mode, length=length, master=self
        )
        marker_mode = SlaveStraightLine.get_opposite_mode(self.mode)
        self._start_mark_line = MarkLine(
            simple_name="start_mark_line",
            mode=marker_mode,
            master=self,
            placement="start",
        )
        self._end_mark_line = MarkLine(
            simple_name="end_mark_line",
            mode=marker_mode,
            master=self,
            placement="end",
            show=False,
        )

    @property
    def is_vertical(self) -> bool:
        if self.mode in ["v", "vertical"]:
            return True
        return False

    @property
    def is_horizontal(self) -> bool:
        if self.mode in ["h", "horizontal"]:
            return True
        return False

    @property
    def straight_line(self) -> SlaveStraightLine:
        return self._straight_line

    @property
    def start_mark_line(self) -> MarkLine:
        return self._start_mark_line

    @property
    def end_mark_line(self) -> MarkLine:
        return self._end_mark_line

    @property
    def mode(self) -> HorizontalVertical:
        return self.straight_line.mode

    @property
    def length(self) -> float:
        return self.straight_line.length

    @length.setter
    def length(self, value: ConvertibleToFloat) -> None:
        self.straight_line.length = value  # type: ignore

    def get_slave_margin(self, slave: SlaveStraightLine, margin: MarginType) -> float:
        check_type(
            margin,
            "MarginType",
            class_name="self.__class__.__name__",
            method_name="get_slave_margin",
            argument_name="margin",
        )
        return 0

    def _get_max_markline_length(self) -> float:
        return max([self.start_mark_line.length, self.end_mark_line.length])

    def _get_slave_x(self, slave: SlaveStraightLine) -> float:
        max_markline_length = self._get_max_markline_length()
        if self.is_horizontal:
            if slave == self.end_mark_line:
                return self.get_relative_x2()
            else:
                return self.relative_x
        else:
            if slave == self.straight_line:
                return self.relative_x + max_markline_length / 2
            elif slave.length == max_markline_length:
                return self.relative_x
            else:
                return self.relative_x + (max_markline_length - slave.length) / 2

    def _get_slave_y(self, slave: SlaveStraightLine) -> float:
        max_markline_length = self._get_max_markline_length()
        if self.is_vertical:
            if slave == self.end_mark_line:
                return self.get_relative_y2()
            return self.relative_y
        else:
            if slave == self.straight_line:
                return self.relative_y + max_markline_length / 2
            elif slave.length == max_markline_length:
                return self.relative_y
            else:
                return self.relative_y + (max_markline_length - slave.length) / 2

    def get_slave_position(
        self, slave: SlaveStraightLine, position: PositionType
    ) -> float:
        check_type(
            position,
            "PositionType",
            class_name="self.__class__.__name__",
            method_name="get_slave_position",
            argument_name="position",
        )
        if position == "x":
            return self._get_slave_x(slave)
        else:
            return self._get_slave_y(slave)

    def set_straight_line_relative_y(self, val: Union[int, float]) -> None:
        if self.is_vertical:
            raise NotImplementedError  # pragma: no cover
        self.relative_y = val - self._get_max_markline_length() / 2

    def set_straight_line_relative_x(self, val: Union[int, float]) -> None:
        if self.is_horizontal:
            raise NotImplementedError  # pragma: no cover
        self.relative_x = val - self._get_max_markline_length() / 2


class HorizontalLineSegment(LineSegment):
    def __init__(self, length: ConvertibleToFloat, *args: Any, **kwargs: Any):
        super().__init__(mode="horizontal", length=length, *args, **kwargs)  # type: ignore

    def get_relative_x2(self) -> float:
        return self.relative_x + self.length

    def get_relative_y2(self) -> float:
        return self.relative_y + self._get_max_markline_length()

    def draw(self, pdf: Pdf) -> None:
        self.start_mark_line.draw(pdf)
        self.straight_line.draw(pdf)
        self.end_mark_line.draw(pdf)


class VerticalLineSegment(LineSegment):
    def __init__(self, length: ConvertibleToFloat, *args: Any, **kwargs: Any):
        super().__init__(mode="vertical", length=length, *args, **kwargs)  # type: ignore

    def get_relative_x2(self) -> float:
        return self.relative_x + max(
            [ml.get_width() for ml in [self.start_mark_line, self.end_mark_line]]
        )

    def get_relative_y2(self) -> float:
        return self.relative_y + self.length

    def draw(self, pdf: Pdf) -> None:
        self.start_mark_line.draw(pdf)
        self.straight_line.draw(pdf)
        self.end_mark_line.draw(pdf)


class AbstractSegmentedLine(DrawObjectContainer):
    def __init__(self, lengths: list[ConvertibleToFloat], *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._make_segments(lengths)

    @property
    def lengths(self) -> None:
        raise AttributeError("use get_lengths instead")

    @lengths.setter
    def lengths(self, lengths: Any) -> None:
        raise SegmentedLineLengthsCannotBeSetError(
            "lengths are not settable after initialization."
        )

    def _check_segment_margins(self) -> None:
        for seg in self.segments:
            if seg.margins != (0, 0, 0, 0):
                raise SegmentedLineSegmentHasMarginsError()

    @property
    def segments(self) -> list[LineSegment]:
        return cast(list[LineSegment], self.get_draw_objects())

    @abstractmethod
    def _make_segments(self, lengths: list[ConvertibleToFloat]) -> None:
        """private method for making segments"""

    @property
    def is_vertical(self) -> bool:
        return self.segments[0].is_vertical

    @property
    def is_horizontal(self) -> bool:
        return self.segments[0].is_horizontal

    def set_straight_line_relative_y(self, val: Union[float, int]) -> None:
        if self.is_horizontal:
            delta = val - self.segments[0].straight_line.relative_y
            self.relative_y += delta
        else:
            raise NotImplementedError  # pragma: no cover

    def set_straight_line_relative_x(self, val: Union[float, int]) -> None:
        if self.is_vertical:
            delta = val - self.segments[0].straight_line.relative_x
            self.relative_x += delta
        else:
            raise NotImplementedError  # pragma: no cover

    def get_lengths(self) -> list[float]:
        return [seg.length for seg in self.segments]


class HorizontalSegmentedLine(AbstractSegmentedLine, DrawObjectRow):
    def _make_segments(self, lengths: list[ConvertibleToFloat]) -> None:
        if lengths:
            for length in lengths:
                self.add_draw_object(HorizontalLineSegment(length))
            self.segments[-1].end_mark_line.show = True

    def _align_segments(self) -> None:
        reference_segment = max(self.segments, key=lambda seg: seg.get_height())
        for segment in self.segments:
            if segment != reference_segment:
                segment.set_straight_line_relative_y(
                    reference_segment.straight_line.relative_y
                )

    def draw(self, pdf: Pdf) -> None:
        self._check_segment_margins()
        self._align_segments()
        super().draw(pdf)


class VerticalSegmentedLine(AbstractSegmentedLine, DrawObjectColumn):
    def _make_segments(self, lengths: list[ConvertibleToFloat]) -> None:
        if lengths:
            for length in lengths:
                self.add_draw_object(VerticalLineSegment(length))
            self.segments[-1].end_mark_line.show = True

    def _align_segments(self) -> None:
        reference_segment = max(self.segments, key=lambda seg: seg.get_width())
        for segment in self.segments:
            if segment != reference_segment:
                segment.set_straight_line_relative_x(
                    reference_segment.straight_line.relative_x
                )

    def draw(self, pdf: Pdf) -> None:
        self._check_segment_margins()
        self._align_segments()
        super().draw(pdf)



from musurgia.trees.timelinetree import TimelineDuration, TimelineTree
import itertools
from fractions import Fraction
from typing import Union, Optional, List, Callable, Any, cast, Sequence, TypeVar


from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.matrix.matrix import (
    PermutationOrderMatrix,
    PermutationOrderMatrixGenerator,
)
from musurgia.musurgia_exceptions import (
    FractalTimelineTreeHasChildrenError,
    FractalTimelineTreeNoneRootCannotSetMainPermutationOrderError,
    PermutationIndexCalculaterNoParentIndexError,
    FractalTimelineTreePermutationIndexError,
    FractalTimelineTreeSetMainPermutationOrderFirstError,
    FractalTimelineTreeMergeWrongValuesError,
    FractalTimelineTreeHasNoChildrenError,
)
from musurgia.musurgia_types import (
    ConvertibleToFraction,
    FractalTreeReduceChildrenMode,
    convert_to_fraction,
    MatrixIndex,
    PermutationOrder,
    check_type,
    PositiveInteger,
    check_matrix_index_values,
    create_error_message,
)
from musurgia.permutation.permutation import permute

__all__ = ["FractalTimelineTree"]

T = TypeVar("T", bound="FractalTimelineTree")


def node_info(node: "FractalTimelineTree") -> str:
    return f"{node.get_fractal_order()}: {node.get_permutation_index()}: {round(float(node.get_value()), 2)}"


class PermutationIndexCalculater:
    def __init__(
        self,
        size: PositiveInteger,
        parent_index: Optional[MatrixIndex] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._size: PositiveInteger
        self._parent_index: Optional[MatrixIndex] = None

        self.size = size
        self.parent_index = parent_index

    @property
    def size(self) -> PositiveInteger:
        return self._size

    @size.setter
    def size(self, value: PositiveInteger) -> None:
        check_type(
            value,
            "PositiveInteger",
            class_name=self.__class__.__name__,
            property_name="size",
        )
        self._size = value

    @property
    def parent_index(self) -> Optional[MatrixIndex]:
        return self._parent_index

    @parent_index.setter
    def parent_index(self, value: MatrixIndex) -> None:
        if value is not None:
            check_type(
                value,
                "MatrixIndex",
                class_name=self.__class__.__name__,
                property_name="parent_index",
            )
        self._parent_index = value

    def get_index(self, column_number: PositiveInteger) -> MatrixIndex:
        if self.parent_index is None:
            raise PermutationIndexCalculaterNoParentIndexError
        check_type(
            column_number,
            "PositiveInteger",
            class_name=self.__class__.__name__,
            method_name="get_index",
            argument_name="column_number",
        )
        if column_number > self.size:
            raise ValueError(
                create_error_message(
                    class_name=self.__class__.__name__,
                    method_name="get_index",
                    argument_name="column_number",
                    message=f"Column number {column_number} cannot be less than size {self.size}",
                )
            )
        r = sum(self.parent_index) % self.size
        if r == 0:
            r = self.size
        return r, column_number


class FractalTimelineTree(TimelineTree):
    def __init__(
        self,
        proportions: Sequence[ConvertibleToFraction],
        main_permutation_order: Optional[PermutationOrder] = None,
        permutation_index: Optional[MatrixIndex] = None,
        fertile: bool = True,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._permutation_order_matrix: Optional[PermutationOrderMatrix] = None
        self._value: Fraction
        self._proportions: Sequence[ConvertibleToFraction] = []
        self._main_permutation_order: Optional[PermutationOrder] = None
        self._permutation_index: Optional[MatrixIndex] = None
        self._fertile: bool

        self._fractal_order: int = 0
        self._children_fractal_values: list[Fraction]
        self._children_permutation_order_matrices = None
        self._permutation_order: tuple[int, int]

        self.proportions = proportions
        self.main_permutation_order = main_permutation_order
        self.set_permutation_index(permutation_index)
        self.fertile = fertile

        self._pic: PermutationIndexCalculater

    def _calculate_children_fractal_values(self) -> list["Fraction"]:
        return permute(
            [self.get_value() * prop for prop in self.proportions],
            self.get_permutation_order(),
        )

    def _get_children_fractal_values(self) -> list["Fraction"]:
        """
        >>> ft = FractalTimelineTree(duration=TimelineDuration(10), proportions=(1, 2, 3), main_permutation_order=(3, 1, 2), permutation_index=(1, 1))
        >>> ft.add_layer()
        >>> ft._get_children_fractal_values()
        [Fraction(5, 1), Fraction(5, 3), Fraction(10, 3)]
        """
        try:
            return self._children_fractal_values
        except AttributeError:
            self._children_fractal_values = self._calculate_children_fractal_values()
            return self._children_fractal_values

    def _get_merge_lengths(self, size: int, merge_index: int) -> list[int]:
        if size == 1:
            return [self.get_size()]

        lengths = self.get_size() * [1]
        pointer = merge_index
        sliced_lengths = [lengths[:pointer], lengths[pointer:]]

        if not sliced_lengths[0]:
            sliced_lengths = sliced_lengths[1:]

        while len(sliced_lengths) < size and len(sliced_lengths[0]) > 1:
            temp = sliced_lengths[0]
            sliced_lengths[0] = temp[:-1]
            sliced_lengths.insert(1, temp[-1:])

        while len(sliced_lengths) < size and len(sliced_lengths[pointer]) > 1:
            temp = sliced_lengths[pointer]
            sliced_lengths[pointer] = temp[:-1]
            sliced_lengths.insert(pointer + 1, temp[-1:])

        output = [len(x) for x in sliced_lengths]

        return output

    def _get_pic(self) -> PermutationIndexCalculater:
        if self.is_root:
            try:
                return self._pic
            except AttributeError:
                raise FractalTimelineTreeSetMainPermutationOrderFirstError
            # except AttributeError:
            #     self._pic = PermutationIndexCalculater(self.get_permutation_order_matrix().get_size())
            #     return self._pic

        return cast(PermutationIndexCalculater, self.get_root()._get_pic())

    # properties
    @property
    def fertile(self) -> bool:
        return self._fertile

    @fertile.setter
    def fertile(self, val: bool) -> None:
        self._fertile = val

    @property
    def main_permutation_order(self) -> Optional[PermutationOrder]:
        if self.is_root:
            return self._main_permutation_order
        else:
            return cast(FractalTimelineTree, self.get_root()).main_permutation_order

    @main_permutation_order.setter
    def main_permutation_order(self, value: Optional[PermutationOrder]) -> None:
        if self._get_children():
            raise FractalTimelineTreeHasChildrenError
        if value is not None:
            if not self.is_root:
                raise FractalTimelineTreeNoneRootCannotSetMainPermutationOrderError
            check_type(
                value,
                "PermutationOrder",
                class_name=self.__class__.__name__,
                property_name="main_permutation_order",
            )
            self._permutation_order_matrix = PermutationOrderMatrixGenerator(
                main_permutation_order=value
            ).generate_permutation_order_matrix()
            self._pic = PermutationIndexCalculater(
                self.get_permutation_order_matrix().get_size()
            )
        else:
            self._permutation_order_matrix = None
        self._main_permutation_order = value

    @property
    def proportions(self) -> Sequence[ConvertibleToFraction]:
        return self._proportions

    @proportions.setter
    def proportions(self, values: Sequence[ConvertibleToFraction]) -> None:
        converted_values: List[Fraction] = [
            convert_to_fraction(val) for val in list(values)
        ]
        total = sum(converted_values)
        self._proportions = [Fraction(value, total) for value in converted_values]

    # public methods
    def add_layer(
        self, *conditions: Optional[Callable[["FractalTimelineTree"], bool]]
    ) -> None:
        """
        >>> ft = FractalTimelineTree(duration=TimelineDuration(10), proportions=(1, 2, 3), main_permutation_order=(3, 1, 2), permutation_index=(1, 1))
        >>> ft.add_layer()
        >>> ft.get_leaves(key=lambda leaf: leaf.get_fractal_order())
        [3, 1, 2]
        >>> ft.get_leaves(key=lambda leaf: leaf.get_permutation_index())
        [(2, 1), (2, 2), (2, 3)]
        >>> ft.get_leaves(key=lambda leaf: round(float(leaf.get_value() ), 2))
        [5.0, 1.67, 3.33]
        >>> ft.add_layer()
        >>> ft.get_leaves(key=lambda leaf: leaf.get_fractal_order())
        [[1, 2, 3], [3, 1, 2], [2, 3, 1]]
        >>> ft.get_leaves(key=lambda leaf: round(float(leaf.get_value() ), 2))
        [[0.83, 1.67, 2.5], [0.83, 0.28, 0.56], [1.11, 1.67, 0.56]]


        >>> ft = FractalTimelineTree(duration=TimelineDuration(10), proportions=(1, 2, 3), main_permutation_order=(3, 1, 2), permutation_index=(1, 1))
        >>> ft.add_layer()
        >>> ft.get_children()[0].add_layer()
        >>> ft.get_leaves(key=lambda leaf: leaf.get_fractal_order())
        [[1, 2, 3], 1, 2]
        >>> ft.get_leaves(key=lambda leaf: round(float(leaf.get_value() ), 2))
        [[0.83, 1.67, 2.5], 1.67, 3.33]
        """

        leaves = list(self.iterate_leaves())

        if conditions is not None:
            for leaf in leaves:
                for condition in conditions:
                    if (
                        cast(Callable[["FractalTimelineTree"], bool], condition)(leaf)
                        is False
                    ):
                        leaf.fertile = False
                        break

        for leaf in leaves:
            if leaf.fertile is True:
                for i in range(leaf.get_size()):
                    value = leaf._get_children_fractal_values()[i]
                    new_node = self.__class__(
                        duration=TimelineDuration(value),
                        proportions=self.get_root().proportions,
                        permutation_index=None,
                    )
                    leaf.add_child(new_node)
                    new_node.calculate_permutation_index()
                    new_node._fractal_order = leaf.get_children_fractal_orders()[i]
            else:
                pass

    def calculate_permutation_index(self: T) -> None:
        if self.is_root:
            raise FractalTimelineTreePermutationIndexError(
                f"{self.__class__.__name__}:calculate_permutation_index: Set permutation_index of root"
            )
        pic = self._get_pic()
        parent = cast(T, self.up)
        pic.parent_index = parent.get_permutation_index()
        self._permutation_index = pic.get_index(parent._get_children().index(self) + 1)

    def generate_children(
        self,
        number_of_children: Union[int, tuple[int, ...], tuple[tuple[int, ...], ...]],
        reduce_mode: FractalTreeReduceChildrenMode = "backwards",
        merge_index: int = 0,
    ) -> None:
        """
        :param number_of_children:
        :param mode:
        :param merge_index:

        >>> ft = FractalTimelineTree(duration=TimelineDuration(10), proportions=(1, 2, 3), main_permutation_order=(3, 1, 2), permutation_index=(1, 1))
        >>> ft.generate_children(number_of_children=((1, 3), 2, (1, (1, 3), 3)))
        >>> print(ft.get_tree_representation(key=lambda node: str(node.get_fractal_order())))
        └── 0
            ├── 3
            │   ├── 1
            │   │   └── 3
            │   ├── 2
            │   │   ├── 2
            │   │   │   └── 3
            │   │   └── 3
            │   │       ├── 2
            │   │       ├── 3
            │   │       └── 1
            │   └── 3
            │       ├── 3
            │       ├── 1
            │       └── 2
            ├── 1
            │   ├── 3
            │   │   ├── 3
            │   │   ├── 1
            │   │   └── 2
            │   └── 2
            │       └── 3
            └── 2
                ├── 2
                └── 3
        <BLANKLINE>
        """
        # check_generate_children_mode(reduce_mode)
        # this error must be moved to add_layer()
        if self._get_children():
            raise ValueError(
                f"FractalTimelineTree.generate_children: node has already children: {[ch.get_value() for ch in self._get_children()]}"
            )

        if isinstance(number_of_children, int):
            if number_of_children > self.get_size():
                raise ValueError(
                    f"generate_children.number_of_children {number_of_children} can not be a greater than size {self.get_size()}"
                )
            if number_of_children < 0:
                raise ValueError(
                    "generate_children.number_of_children {} must be a positive int".format(
                        number_of_children
                    )
                )
            elif number_of_children == 0:
                pass
            else:
                self.add_layer()
                self.reduce_children_by_size(
                    size=number_of_children, mode=reduce_mode, merge_index=merge_index
                )

        elif isinstance(number_of_children, tuple):
            self.generate_children(
                len(number_of_children),
                reduce_mode=reduce_mode,
                merge_index=merge_index,
            )

            for index, child in enumerate(self._get_children()):
                if reduce_mode == "backwards":
                    number_of_grand_children = number_of_children[
                        child.get_fractal_order()
                        - child.get_size()
                        + len(number_of_children)
                        - 1
                    ]
                else:
                    number_of_grand_children = number_of_children[index]
                child.generate_children(
                    number_of_grand_children,
                    reduce_mode=reduce_mode,
                    merge_index=merge_index,
                )

        else:
            raise TypeError(
                "generate_children.number_of_children must be of type int or tuple"
            )

    def get_children_fractal_orders(self) -> list[int]:
        if self.is_root:
            if self.main_permutation_order is None:
                raise FractalTimelineTreeSetMainPermutationOrderFirstError(
                    create_error_message(
                        message="Set main_permutation_order first",
                        class_name=self.__class__.__name__,
                        method_name="get_children_fractal_orders",
                    )
                )
            else:
                return permute(
                    list(range(1, self.get_size() + 1)), self.main_permutation_order
                )
        return permute(
            list(range(1, self.get_size() + 1)), self.get_permutation_order()
        )

    def get_fractal_order(self) -> int:
        """
        :return:

        >>> ft = FractalTimelineTree(duration=TimelineDuration(10), proportions=(1, 2, 3), main_permutation_order=(3, 1, 2), permutation_index=(1, 1))
        >>> ft.add_layer()
        >>> [node.get_fractal_order() for node in ft.traverse()]
        [0, 3, 1, 2]
        """
        return self._fractal_order

    # def get_layer(self, layer: int, key: Optional[Callable[['FractalTree'], Any]] = None) -> Any:
    # """
    # :param layer:
    # :param key:
    # :return:

    # >>> ft = FractalTree(10, (1, 2, 3), (3, 1, 2), (1, 1))
    # >>> ft.add_layer()
    # >>> for i in range(3):
    # ...     ft.add_layer(lambda n: True if n.get_fractal_order() > 1 else False)
    # >>> print(ft.get_layer(0, key=lambda node: node.get_fractal_order()))
    # None
    # >>> ft.get_layer(1, key=lambda node: node.get_fractal_order())
    # [3, 1, 2]
    # >>> ft.get_layer(2, key=lambda node: node.get_fractal_order())
    # [[1, 2, 3], 1, [2, 3, 1]]
    # >>> ft.get_layer(3, key=lambda node: node.get_fractal_order())
    # [[1, [1, 2, 3], [3, 1, 2]], 1, [[1, 2, 3], [3, 1, 2], 1]]
    # >>> ft.get_layer(4, key=lambda node: node.get_fractal_order())
    # [[1, [1, [3, 1, 2], [2, 3, 1]], [[2, 3, 1], 1, [3, 1, 2]]], 1, [[1, [1, 2, 3], [3, 1, 2]], [[3, 1, 2], 1, [1, 2, 3]], 1]]
    # """
    # if layer > self.get_root().get_number_of_layers():
    #     raise ValueError(f'FractalTree.get_layer: max layer number={self.get_number_of_layers()}')
    # else:
    #     if layer == 0:
    #         return cast('FractalTree', self.get_self_with_key(key))
    #     else:
    #         if self.is_leaf:
    #             return self.get_layer(layer=layer - 1, key=key)
    #         output = []
    #         for child in self.get_children():
    #             if child.get_farthest_leaf().get_distance() == 1:
    #                 output.append(child.get_self_with_key(key))
    #             else:
    #                 output.append(child.get_layer(layer - 1, key))
    #         return output

    def get_permutation_order(self) -> tuple[int, int]:
        try:
            return self._permutation_order
        except AttributeError:
            self._permutation_order = self.get_permutation_order_matrix().get_element(
                cast(MatrixIndex, self.get_permutation_index())
            )
            return self._permutation_order

    def get_permutation_index(self) -> Optional[MatrixIndex]:
        return self._permutation_index

    def get_permutation_order_matrix(self) -> PermutationOrderMatrix:
        if self.is_root:
            if self._permutation_order_matrix is None:
                raise FractalTimelineTreeSetMainPermutationOrderFirstError(
                    create_error_message(
                        message=FractalTimelineTreeSetMainPermutationOrderFirstError.msg,
                        class_name=self.__class__.__name__,
                        method_name="get_permutation_order_matrix",
                    )
                )
            return self._permutation_order_matrix
        else:
            return cast(
                FractalTimelineTree, self.get_root()
            ).get_permutation_order_matrix()

    def get_size(self) -> int:
        """
        >>> ft = FractalTimelineTree(duration=TimelineDuration(10), proportions=(1, 2, 3), main_permutation_order=(3, 1, 2))
        >>> ft.get_size()
        3
        """
        return len(self.proportions)

    def merge_children(self, *lengths: int) -> None:
        """

        :param lengths:
        :return:

        >>> ft = FractalTimelineTree(proportions=(1, 2, 3, 4, 5), main_permutation_order=(3, 5, 1, 2, 4), duration=TimelineDuration(10), permutation_index=(1, 1))
        >>> ft.add_layer()
        >>> print(ft.get_tree_representation(node_info))
        └── 0: (1, 1): 10.0
            ├── 3: (2, 1): 2.0
            ├── 5: (2, 2): 3.33
            ├── 1: (2, 3): 0.67
            ├── 2: (2, 4): 1.33
            └── 4: (2, 5): 2.67
        <BLANKLINE>
        >>> ft.merge_children(1, 2, 2)
        >>> print(ft.get_tree_representation(node_info))
        └── 0: (1, 1): 10.0
            ├── 3: (2, 1): 2.0
            ├── 5: (2, 2): 4.0
            └── 2: (2, 4): 4.0
        <BLANKLINE>
        """
        children = self._get_children()
        if not children:
            raise FractalTimelineTreeHasNoChildrenError(
                "FractalTimelineTree.merge_children: There are no children to be merged"
            )
        if sum(lengths) != len(children):
            raise FractalTimelineTreeMergeWrongValuesError(
                f"FractalTimelineTree.merge_children: Sum of lengths {sum(lengths)} must be the same as length of children {len(children)}"
            )

        def _merge(nodes: list["FractalTimelineTree"]) -> None:
            node_values = [node.get_value() for node in nodes]
            new_value = sum(node_values)
            for node in nodes[1:]:
                self.remove(node)
            nodes[0].update_value(new_value)

        iter_children = iter(children)
        chunks = [list(itertools.islice(iter_children, l)) for l in lengths]

        for chunk in chunks:
            _merge(chunk)

    def reduce_children_by_condition(
        self, condition: Callable[["FractalTimelineTree"], bool]
    ) -> None:
        if not self._get_children():
            raise FractalTimelineTreeHasNoChildrenError(
                f"{self} has no children to be reduced"
            )
        for child in [child for child in self._get_children() if condition(child)]:
            self.remove(child)
            del child
        reduced_value = sum([child.get_value() for child in self._get_children()])
        factor = self.get_value() / reduced_value
        for child in self._get_children():
            new_value = child.get_value() * factor
            child.update_value(new_value)

        self._children_fractal_values = [
            child.get_value() for child in self._get_children()
        ]

    def reduce_children_by_size(
        self,
        size: int,
        mode: FractalTreeReduceChildrenMode = "backwards",
        merge_index: Optional[int] = None,
    ) -> None:
        check_type(
            mode,
            "FractalTreeReduceChildrenMode",
            class_name=self.__class__.__name__,
            method_name="reduce_children_by_size",
            argument_name="mode",
        )

        if size > self.get_size() or size < 0:
            raise ValueError(
                f"reduce_children_by_size.size {size} must be a positive int not greater than {self.get_size()}"
            )
        if size == 0:
            pass
        else:
            if mode == "backwards":
                self.reduce_children_by_condition(
                    lambda child: child.get_fractal_order() < self.get_size() - size + 1
                )
            elif mode == "forwards":
                self.reduce_children_by_condition(
                    lambda child: child.get_fractal_order() > size
                )
            elif mode == "sieve":
                if size == 1:
                    self.reduce_children_by_condition(
                        condition=lambda child: child.get_fractal_order() not in [1]
                    )
                else:
                    ap = ArithmeticProgression(a1=1, an=self.get_size(), n=size)
                    selection = [int(round(x)) for x in ap]
                    self.reduce_children_by_condition(
                        condition=lambda child: child.get_fractal_order()
                        not in selection
                    )
            elif mode == "merge":
                if merge_index is None:
                    raise ValueError(
                        "reduce_children.merge_index must be set for mode merge"
                    )
                if merge_index > self.get_size() - 1:
                    raise ValueError(
                        f"reduce_children_by_size.merge_index {merge_index} must be a positive int not greater than {self.get_size() - 1}"
                    )
                merge_lengths = self._get_merge_lengths(size, merge_index)
                self.merge_children(*merge_lengths)

    def set_permutation_index(self, index: Optional[MatrixIndex]) -> None:
        if index is not None:
            check_type(
                index,
                "MatrixIndex",
                class_name=self.__class__.__name__,
                method_name="set_permutation_index",
                argument_name="index",
            )
            size = self.get_permutation_order_matrix().get_size()
            check_matrix_index_values(index, size, size)
        self._permutation_index = index

    def split(self: "T", *proportions: Any) -> list["T"]:
        if self._get_children():
            raise FractalTimelineTreeHasChildrenError

        if hasattr(proportions[0], "__iter__"):
            proportions = proportions[0]

        proportions_list = [Fraction(prop) for prop in proportions]

        for prop in proportions_list:
            duration = self.get_duration() * prop / sum(proportions_list)
            new_node = self.__class__(
                duration=duration,
                proportions=self.get_root().proportions,
                permutation_index=None,
            )
            new_node._fractal_order = self.get_fractal_order()
            self.add_child(new_node)
            new_node.calculate_permutation_index()

        return self._get_children()



from unittest import TestCase

from fractions import Fraction

from musicscore.util import dToX, xToD

from musurgia.quantize import get_quantized_values, find_best_quantized_values


class Test(TestCase):
    def test_1(self):
        input = [0.2, 0.333, 0.6, 0.99, 0.1, 0.5]
        example = get_quantized_values(input, 0.6)
        result = [
            Fraction(0, 1),
            Fraction(3, 5),
            Fraction(3, 5),
            Fraction(6, 5),
            Fraction(0, 1),
            Fraction(0, 1),
        ]
        self.assertEqual(result, example)

    def test_2(self):
        midis = [60, 61.5, 58.5, 56.5, 63]
        intervals = xToD(midis)
        quantized_intervals = get_quantized_values(intervals, 1)
        quantized_midis = [
            int(x) for x in dToX(quantized_intervals, first_element=midis[0])
        ]
        result = [60, 61, 58, 57, 63]
        self.assertEqual(result, quantized_midis)

    def test_find_best_unit_without_check_sum(self):
        values = [0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]
        units = [0.5, 0.7, 0.2]
        actual = find_best_quantized_values(
            values=values, list_of_grids=units, check_sum=False
        )
        expected = [
            Fraction(1, 5),
            Fraction(1, 5),
            Fraction(2, 5),
            Fraction(3, 5),
            Fraction(3, 5),
            Fraction(3, 5),
            Fraction(4, 5),
        ]
        self.assertEqual(expected, actual)

    def test_find_best_unit_with_check_sum(self):
        values = [Fraction(1, 7), Fraction(3, 7), Fraction(5, 7)]
        units = [Fraction(1, 2), Fraction(1, 7), Fraction(1, 5)]
        actual = find_best_quantized_values(
            values=values, list_of_grids=units, check_sum=True
        )
        expected = [Fraction(1, 7), Fraction(3, 7), Fraction(5, 7)]

        self.assertEqual(expected, actual)

    def test_find_best_unit_with_check_sum_2(self):
        values = [1, 2, 3, 4, 5]
        units = [Fraction(1, 2), Fraction(1, 7), Fraction(1, 5)]
        actual = find_best_quantized_values(
            values=values, list_of_grids=units, check_sum=True
        )
        expected = [
            Fraction(1, 1),
            Fraction(2, 1),
            Fraction(3, 1),
            Fraction(4, 1),
            Fraction(5, 1),
        ]

        self.assertEqual(expected, actual)



from .drawobject import *
from .font import *
from .labeled import *
from .line import *
from .margined import *
from .pdf import *
from .pdf_tools import *
from .positioned import *
from .rowcolumn import *
from .text import *
from .ruler import *



from unittest import TestCase

from musicscore.midi import Midi

from musurgia.chordfactory.midigenerators import (
    RandomMidiGenerator,
    RelativeMidiGenerator,
)


class RandomMidiGeneratorTestCase(TestCase):
    def setUp(self):
        self.rmg = RandomMidiGenerator(
            pool=[60, 61, 62.5, 64, 52], periodicity=2, seed=10
        )

    def test_generate_random_midis(self):
        random_midis = [next(self.rmg) for _ in range(10)]
        for midi in random_midis:
            self.assertTrue(isinstance(midi, Midi))
        self.assertListEqual(
            [m.value for m in random_midis], [52, 60, 64, 52, 60, 61, 64, 62.5, 61, 60]
        )


class RelativeMidiGeneratorTestCase(TestCase):
    def setUp(self):
        self.remg = RelativeMidiGenerator(
            value_range=[44, 54], proportions=[2, 4, 1, 3], directions=[-1, -1, 1, 1]
        )

    def test_init_relative_midi_generator(self):
        expected = [54.0, 51.0, 44.0, 46.0, 51.0]
        self.assertListEqual([midi.value for midi in self.remg], expected)



from pprint import pprint  # noqa
from typing import Any

from musurgia.musurgia_exceptions import (
    PermutationOrderTypeError,
    PermutationOrderValueError,
)
from musurgia.musurgia_types import (
    check_type,
    PermutationOrder,
    check_permutation_order_values,
    MatrixData,
)


def permute(input_list: list[Any], permutation_order: PermutationOrder) -> list[Any]:
    """
    Permutes a list of values by reference to another list named permutation_order.
    :param input_list: A list of values to permute
    :param permutation_order: A tuple consisting of all integers between 1 and lenght of input_list - 1 without duplications.
    :return: permuted list of values

    >>> permute([10, 20, 30, 40], (3, 2, 4, 1))
    [30, 20, 40, 10]
    """
    check_type(input_list, list, function_name="permute", argument_name="input_list")
    try:
        check_type(
            permutation_order,
            "PermutationOrder",
            function_name="permute",
            argument_name="permutation_order",
        )
    except TypeError as err:
        raise PermutationOrderTypeError(err)

    try:
        check_permutation_order_values(permutation_order, len(input_list))
    except ValueError as err:
        raise PermutationOrderValueError(err)

    return [input_list[m - 1] for m in permutation_order]


def get_self_permutation_2d(
    permutation_order: tuple[int, ...],
) -> list[tuple[int, ...]]:
    """
    This is a function for applying the `permutation_order` to itself.

    :param permutation_order: A tuple consisting of all integers between 1 and a higher integer

    :return: A list of `permutations_orders`  as a result of applying the `permutation_order` to itself recursively. The
    result has always a length equal to `len(permutation_order)`.
    Each permuted permutation order will be permuted again. If all integers of the original permuation are not in their
    natural order (natural orders=`(1, 2, 3, 4, ...)`) the resulted list will be distinctive. Otherwise there will be duplicates.

    >>> get_self_permutation_2d((3, 1, 2))
    [(3, 1, 2), (2, 3, 1), (1, 2, 3)]

    >>> get_self_permutation_2d((1, 3, 2))
    [(1, 3, 2), (1, 2, 3), (1, 3, 2)]

    >>> get_self_permutation_2d((1, 2, 3))
    [(1, 2, 3), (1, 2, 3), (1, 2, 3)]

    >>> get_self_permutation_2d((3, 1, 4, 2))
    [(3, 1, 4, 2), (4, 3, 2, 1), (2, 4, 1, 3), (1, 2, 3, 4)]

    >>> get_self_permutation_2d((4, 2, 3, 1))
    [(4, 2, 3, 1), (1, 2, 3, 4), (4, 2, 3, 1), (1, 2, 3, 4)]

    """
    output = [permutation_order]

    for i in range(1, len(permutation_order)):
        output.append(tuple(permute(list(output[i - 1]), permutation_order)))

    return output


def get_self_permutation_3d(permutation_order: tuple[int, ...]) -> MatrixData:
    """
    This is a function for applying the `permutation_order` to itself in a higher order compared to :obj:`get_self_permutation_2d`.
    If :obj:`get_self_permutation_2d` is a two dimensional reflexive operation, :obj:`get_self_permutation_3d` is a three
    dimensional one.

    :param permutation_order: A list consisting of all integers between 1 and a higher integer

    :return:

    >>> pprint(get_self_permutation_3d((3, 1, 2)))
    [[(3, 1, 2), (2, 3, 1), (1, 2, 3)],
     [(1, 2, 3), (3, 1, 2), (2, 3, 1)],
     [(2, 3, 1), (1, 2, 3), (3, 1, 2)]]

    >>> pprint(get_self_permutation_3d((1, 3, 2)))
    [[(1, 3, 2), (1, 2, 3), (1, 3, 2)],
     [(1, 3, 2), (1, 3, 2), (1, 2, 3)],
     [(1, 3, 2), (1, 2, 3), (1, 3, 2)]]

    >>> pprint(get_self_permutation_3d((1, 2, 3)))
    [[(1, 2, 3), (1, 2, 3), (1, 2, 3)],
     [(1, 2, 3), (1, 2, 3), (1, 2, 3)],
     [(1, 2, 3), (1, 2, 3), (1, 2, 3)]]

    >>> pprint(get_self_permutation_3d((3, 1, 4, 2)))
    [[(3, 1, 4, 2), (4, 3, 2, 1), (2, 4, 1, 3), (1, 2, 3, 4)],
     [(2, 4, 1, 3), (3, 1, 4, 2), (1, 2, 3, 4), (4, 3, 2, 1)],
     [(1, 2, 3, 4), (2, 4, 1, 3), (4, 3, 2, 1), (3, 1, 4, 2)],
     [(4, 3, 2, 1), (1, 2, 3, 4), (3, 1, 4, 2), (2, 4, 1, 3)]]

    >>> pprint(get_self_permutation_3d((3, 4, 2, 1)))
    [[(3, 4, 2, 1), (2, 1, 4, 3), (4, 3, 1, 2), (1, 2, 3, 4)],
     [(4, 3, 1, 2), (1, 2, 3, 4), (2, 1, 4, 3), (3, 4, 2, 1)],
     [(2, 1, 4, 3), (3, 4, 2, 1), (1, 2, 3, 4), (4, 3, 1, 2)],
     [(1, 2, 3, 4), (4, 3, 1, 2), (3, 4, 2, 1), (2, 1, 4, 3)]]

    >>> pprint(get_self_permutation_3d((4, 2, 3, 1)))
    [[(4, 2, 3, 1), (1, 2, 3, 4), (4, 2, 3, 1), (1, 2, 3, 4)],
     [(1, 2, 3, 4), (1, 2, 3, 4), (4, 2, 3, 1), (4, 2, 3, 1)],
     [(4, 2, 3, 1), (1, 2, 3, 4), (4, 2, 3, 1), (1, 2, 3, 4)],
     [(1, 2, 3, 4), (1, 2, 3, 4), (4, 2, 3, 1), (4, 2, 3, 1)]]
    """

    self_permuted_order = get_self_permutation_2d(permutation_order)
    output = [self_permuted_order]

    for i in range(1, len(self_permuted_order)):
        output.append(permute(list(output[i - 1]), permutation_order))
    return output



import copy
from unittest import TestCase
from musurgia.pdf import HorizontalLineSegment, StraightLine


class TestDeepCopy(TestCase):
    def test_horizontal_line_segment(self):
        hls = HorizontalLineSegment(length=10)
        hls.bottom_margin = 30
        hls.relative_x = 20
        hls.start_mark_line.add_text_label("start", font_size=15, bottom_margin=50)
        hls.end_mark_line.add_text_label(
            "stop", font_size=25, position="below", top_margin=40
        )

        copied = copy.deepcopy(hls)
        assert copied != hls
        assert copied.length == hls.length
        assert copied.mode == hls.mode
        assert copied.straight_line != hls.straight_line
        assert copied.start_mark_line != hls.start_mark_line
        assert copied.end_mark_line != hls.end_mark_line
        assert copied.get_positions() == hls.get_positions()
        assert copied.get_margins() == hls.get_margins()
        start_mark_text_label = hls.start_mark_line.get_text_labels()[0]
        copied_start_mark_text_label = copied.start_mark_line.get_text_labels()[0]
        end_mark_text_label = hls.end_mark_line.get_text_labels()[0]
        copied_end_mark_text_label = copied.end_mark_line.get_text_labels()[0]

        assert start_mark_text_label != copied_start_mark_text_label
        assert end_mark_text_label != copied_end_mark_text_label
        assert (
            start_mark_text_label.get_positions()
            == copied_start_mark_text_label.get_positions()
        )
        assert (
            end_mark_text_label.get_positions()
            == copied_end_mark_text_label.get_positions()
        )
        assert (
            start_mark_text_label.get_margins()
            == copied_start_mark_text_label.get_margins()
        )
        assert (
            end_mark_text_label.get_margins()
            == copied_end_mark_text_label.get_margins()
        )

        assert start_mark_text_label.value == copied_start_mark_text_label.value
        assert end_mark_text_label.value == copied_end_mark_text_label.value

    def test_margins_and_positions(self):
        line = StraightLine(length=20, mode="h")
        line.margins = (10, 20, 30, 40)
        line.positions = (5, 10)

        copied = copy.deepcopy(line)

        line.margins = (1, 2, 3, 4)
        line.positions = (1, 2)

        assert copied.length == 20
        assert copied.mode == "h"
        assert copied.margins == (10, 20, 30, 40)
        assert copied.positions == (5, 10)

        copied.length = 10
        copied.margins = (0, 0, 0, 0)
        copied.positions = (0, 0)
        copied.mode = "v"

        assert copied.length == 10
        assert copied.mode == "v"
        assert copied.margins == (0, 0, 0, 0)
        assert copied.positions == (0, 0)

        assert line.length == 20
        assert line.mode == "h"
        assert line.margins == (1, 2, 3, 4)
        assert line.positions == (1, 2)
